!     Last change:  AA   21 Dec 1998    0:25 am

! Cholesky factorization and symmetrix matrix inversion
! Applied Statistics algoritms, ported to fortran by A.Miller
!-----------------------------------------------------

module symxinv

implicit none

contains

SUBROUTINE subinv(a, nm, b, n, c, nullty, ifault, det)

!       Remark asr 44.1   Appl. Statist. (1982) Vol.31, No.3

!       A revised and enhanced version of
!       algorithm as7, applied statistics, vol.17, 1968.

! N.B. Arguments W & NDIM have been removed, and DET is OPTIONAL.

!       Forms in c() as lower triangle, a generalised inverse
!       of a sub-matrix of the positive semi-definite symmetric
!       matrix a() of order n, stored as lower triangle.
!       c() may co-incide with a().  nullty is returned as the nullity
!       of a().  ifault is returned as 1 if n.lt.1, 2 if the
!       submatrix is not positive semi-definite, 3 if the elements
!       of b are inadmissible, otherwise zero.

!       arguments:-
!       a()     = input, the symmetric matrix to be inverted, stored in
!                 lower triangular form
!       nm      = input, the order of a
!       n       = input, order of the sub-matrix to be inverted
!       b       = input, an array containing the numbers of rows and
!                 columns of a() to be included.
!       c()     = output, the inverse of a (a generalized inverse if c is
!                 singular), also stored in lower triangular.
!                 c and a may occupy the same locations.
!       nullty  = output, the rank deficiency of a.
!       ifault  = output, error indicator
!                       = 1 if n < 1
!                       = 2 if the submatrix of a is not +ve semi-definite
!                       = 3 if elements of b are inadmissible
!                       = 0 otherwise

!     Auxiliary routine required: SUBCHL from ASR44 (see AS6)

!***************************************************************************

IMPLICIT NONE
INTEGER, PARAMETER               :: dp = SELECTED_REAL_KIND(15, 60)
REAL (dp), INTENT(IN)            :: a(:)
INTEGER, INTENT(IN)              :: nm, b(:), n
REAL (dp), INTENT(OUT)           :: c(:)
REAL (dp), INTENT(OUT), OPTIONAL :: det
INTEGER, INTENT(OUT)             :: nullty, ifault

  !  INTERFACE
  !  SUBROUTINE subchl (a, b, n, u, nullty, ifault, det)
  !    IMPLICIT NONE
  !    INTEGER, PARAMETER               :: dp = SELECTED_REAL_KIND(15, 60)
  !    REAL (dp), INTENT(IN)            :: a(:)
  !    REAL (dp), INTENT(OUT)           :: u(:)
  !    REAL (dp), INTENT(OUT), OPTIONAL :: det
  !    INTEGER, INTENT(IN)              :: b(:), n
  !    INTEGER, INTENT(OUT)             :: nullty, ifault
  !  END SUBROUTINE subchl
  !END INTERFACE
  !
! Local variables
REAL (dp)              :: w(n), x
REAL (dp), PARAMETER   :: zero = 0.0_dp, one = 1.0_dp
INTEGER                :: i, icol, irow, j, jcol, k, l, mdiag, ndiag, nn, nrow

ifault = 3
IF (n > nm) RETURN
IF (b(1) < 1 .OR. b(1) > nm - n + 1) RETURN
IF (n == 1) GO TO 19
DO i = 2, n
  IF (b(i) <= b(i-1) .OR. b(i) > nm - n + i) RETURN
END DO
19 nrow = n
ifault = 1
IF(nrow <= 0) GO TO 100
ifault = 0

!       Cholesky factorization of a, result in c

IF (PRESENT(det)) THEN
  CALL subchl(a, b, nrow, c, nullty, ifault, det)
ELSE
  CALL subchl(a, b, nrow, c, nullty, ifault )
END IF
IF(ifault /= 0) GO TO 100

!       invert c & form the product (cinv)'*cinv, where cinv is the inverse
!       of c, row by row starting with the last row.
!       irow = the row number, ndiag = location of last element in the row.

nn = nrow*(nrow + 1)/2
irow = nrow
ndiag = nn
DO irow = nrow, 1, -1
  IF(c(ndiag) == zero) GO TO 60
  l = ndiag
  DO i = irow, nrow
    w(i) = c(l)
    l = l + i
  END DO
  icol = nrow
  jcol = nn
  mdiag = nn

  DO
    l = jcol
    x = zero
    IF(icol == irow) x = one / w(irow)
    k = nrow
    DO
      IF(k == irow) EXIT
      x = x - w(k)*c(l)
      k = k - 1
      l = l - 1
      IF(l > mdiag) l = l - k + 1
    END DO

    c(l) = x / w(irow)
    IF(icol == irow) GO TO 80
    mdiag = mdiag - icol
    icol = icol - 1
    jcol = jcol - 1
  END DO

  60 l = ndiag
  DO j = irow, nrow
    c(l) = zero
    l = l + j
  END DO

  80 ndiag = ndiag - irow
END DO

100 RETURN
END SUBROUTINE subinv

SUBROUTINE subchl (a, b, n, u, nullty, ifault, det)

!   REMARK ASR 44  APPL. STATIST. (1982) VOL. 31, NO. 3

!   A revised and enhanced version of
!   ALGORITHM AS 6  APPL. STATIST. (1968) VOL. 17, NO. 2

! N.B. Argument NDIM has been removed, and DET is an OPTIONAL argument

!   Given a symmetric matrix of order N as lower triangle in A(),
!   calculates an upper triangle, U(), such that U'U = the sub-matrix
!   of A whose rows and columns are specified in the integer array B().
!   U() may coincide with A().   A() must be +ve semi-definite.
!   ETA is set to multiplying factor determining effective zero for a pivot.
!   NULLTY is returned as number of effective zero pivots.
!   IFAULT is returned as 1 if N <= 0, 2 if A() is not +ve semi-
!   definite, otherwise 0 is returned.

IMPLICIT NONE
INTEGER, PARAMETER               :: dp = SELECTED_REAL_KIND(15, 60)
REAL (dp), INTENT(IN)            :: a(:)
REAL (dp), INTENT(OUT)           :: u(:)
REAL (dp), INTENT(OUT), OPTIONAL :: det
INTEGER, INTENT(IN)              :: b(:), n
INTEGER, INTENT(OUT)             :: nullty, ifault

!     Local variables

REAL (dp) :: w, eta2, x
INTEGER   :: i, icol, ii, ij, irow, j, jj, k, kk, l, m

!     The value of ETA below will depend upon the word length of the
!     computer being used.

REAL (dp), PARAMETER :: eta = 1.E-14_dp, zero = 0.0_dp, one = 1.0_dp

ifault = 1
IF (n <= 0) GO TO 90
ifault = 2
nullty = 0
IF (PRESENT(det)) det = one
j = 1
k = 0
eta2 = eta*eta
DO icol = 1, n
  ij = b(icol)*(b(icol) - 1)/2
  ii = ij + b(icol)
  x = eta2*a(ii)
  l = 0
  DO irow = 1, icol
    kk = b(irow)*(b(irow) + 1)/2
    k = k + 1
    jj = ij + b(irow)
    w = a(jj)
    m = j
    DO i = 1, irow
      l = l + 1
      IF (i == irow) GO TO 20
      w = w - u(l)*u(m)
      m = m + 1
    END DO
    20 IF (irow == icol) GO TO 50
    IF (u(l) == zero) GO TO 30
    u(k) = w/u(l)
    CYCLE
    30 IF (w*w > ABS(x*a(kk))) GO TO 90
    u(k) = zero
  END DO

  50 IF (ABS(w) <= ABS(eta*a(kk))) GO TO 60
  IF (w < zero) GO TO 90
  u(k) = SQRT(w)
  GO TO 70
  60 u(k) = zero
  nullty = nullty + 1
  70 j = j + icol
  IF (PRESENT(det)) det = det*u(k)*u(k)
END DO

ifault = 0
90 RETURN
END SUBROUTINE subchl

END module symxinv
