
! To: Adelchi Azzalini <aa@gwen.stat.unipd.it>
! Subject: toms611.f90
! Date: Fri, 06 Nov 1998 12:12:27 +1100
! From: "Alan.Miller @ vic.cmis.CSIRO.AU  = Alan Miller, Honorary Research Fellow (Retired)" <Alan.Miller@cmis.CSIRO.AU>

!     TOMS algorithm 611, collected algorithms from acm.
!     algorithm appeared in acm-trans. math. software, vol.9, no. 4,
!     dec., 1983, p. 503-524.

! Translation from Fortran 66/77 format to compatibility with ELF90
! by Alan Miller, 2 April 1997
! CSIRO Mathematical & Information Sciences
! Private Bag 10, Clayton South MDC
! Victoria 3169, Australia
! Fax: (+61) 3-9545-8080   Phone: (+61) 3-9545-8036
! Alan.Miller @ vic.cmis.csiro.au
! http://www.ozemail.com.au/~milleraj

MODULE machine_constants
IMPLICIT NONE
INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(15, 60)

CONTAINS

FUNCTION imdcon(k) RESULT(ival)

IMPLICIT NONE
INTEGER, INTENT(IN) :: k
INTEGER             :: ival

!  ***  return integer machine-dependent constants  ***

!     ***  k = 1 means return standard output unit number.   ***
!     ***  k = 2 means return alternate output unit number.  ***
!     ***  k = 3 means return  input unit number.            ***
!          (note -- k = 2, 3 are used only by test programs.)

INTEGER :: mdcon(3) = (/ 6, 0, 5 /)

ival = mdcon(k)

RETURN
!  ***  last card of imdcon follows  ***
END FUNCTION imdcon


FUNCTION rmdcon(k) RESULT(fn_val)

!  ***  return machine dependent constants used by nl2sol  ***

IMPLICIT NONE
INTEGER, INTENT(IN) :: k
REAL (dp)           :: fn_val

!  ***  the constant returned depends upon k...

!  ***        k = 1... smallest pos. eta such that -eta exists.
!  ***        k = 2... square root of eta.
!  ***        k = 3... unit roundoff = smallest pos. no. machep such
!  ***                 that 1 + machep .gt. 1 .and. 1 - machep .lt. 1.
!  ***        k = 4... square root of machep.
!  ***        k = 5... square root of big (see k = 6).
!  ***        k = 6... largest machine no. big such that -big exists.

SELECT CASE (k)
  CASE(1)
    fn_val = TINY(1._dp)
  CASE(2)
    fn_val = SQRT( TINY(1._dp) )
  CASE(3)
    fn_val = EPSILON(1._dp)
  CASE(4)
    fn_val = SQRT( EPSILON(1._dp) )
  CASE(5)
    fn_val = SQRT( HUGE(1._dp) )
  CASE(6)
    fn_val = HUGE(1._dp)
  CASE DEFAULT
    WRITE(*, *) '** Illegal argument, k = ', k, ' to FUNCTION rmdcon **'
END SELECT

RETURN
!  ***  last card of rmdcon follows  ***
END FUNCTION rmdcon

END MODULE machine_constants


MODULE toms611

CONTAINS

SUBROUTINE sumsl (n,d,x,calcf,calcg,iv,liv,lv,v,uiparm,urparm,ufparm)

!  ***  minimize general unconstrained objective function using   ***
!  ***  analytic gradient and hessian approx. from secant update  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: n,liv,lv
INTEGER, INTENT(IN OUT)   :: iv(:),uiparm(:)
REAL (dp), INTENT(IN OUT) :: d(:),x(:),v(:),urparm(:),ufparm
!     dimension v(71 + n*(n+15)/2), uiparm(*), urparm(*)

INTERFACE
  SUBROUTINE calcf(n, x, nf, f, uiparm, urparm, ufparm)
    USE machine_constants
    IMPLICIT NONE
    INTEGER, INTENT(IN)       :: n
    INTEGER, INTENT(IN OUT)   :: nf, uiparm(:)
    REAL (dp), INTENT(IN)     :: x(:)
    REAL (dp), INTENT(OUT)    :: f, urparm(:), ufparm
  END SUBROUTINE calcf

  SUBROUTINE calcg(n, x, nf, g, uiparm, urparm, ufparm)
    USE machine_constants
    IMPLICIT NONE
    INTEGER, INTENT(IN)       :: n, uiparm(:)
    INTEGER, INTENT(IN OUT)   :: nf
    REAL (dp), INTENT(IN)     :: x(:), ufparm
    REAL (dp), INTENT(IN OUT) :: urparm(:)
    REAL (dp), INTENT(OUT)    :: g(:)
  END SUBROUTINE calcg
END INTERFACE

!  ***  purpose  ***

!        this routine interacts with subroutine  sumit  in an attempt
!     to find an n-vector  x*  that minimizes the (unconstrained)
!     objective function computed by  calcf.  (often the  x*  found is
!     a local minimizer rather than a global one.)

!--------------------------  parameter usage  --------------------------

! n........ (input) the number of variables on which  f  depends, i.e.,
!                  the number of components in  x.
! d........ (input/output) a scale vector such that  d(i)*x(i),
!                  i = 1,2,...,n,  are all in comparable units.
!                  d can strongly affect the behavior of sumsl.
!                  finding the best choice of d is generally a trial-
!                  and-error process.  choosing d so that d(i)*x(i)
!                  has about the same value for all i often works well.
!                  the defaults provided by subroutine deflt (see iv
!                  below) require the caller to supply d.
! x........ (input/output) before (initially) calling sumsl, the call-
!                  er should set  x  to an initial guess at  x*.  when
!                  sumsl returns,  x  contains the best point so far
!                  found, i.e., the one that gives the least value so
!                  far seen for  f(x).
! calcf.... (input) a subroutine that, given x, computes f(x).  calcf
!                  must be declared external in the calling program.
!                  it is invoked by
!                       call calcf(n, x, nf, f, uiparm, urparm, ufparm)
!                  when calcf is called, nf is the invocation
!                  count for calcf.  nf is included for possible use
!                  with calcg.  if x is out of bounds (e.g., if it
!                  would cause overflow in computing f(x)), then calcf
!                  should set nf to 0.  this will cause a shorter step
!                  to be attempted.  (if x is in bounds, then calcf
!                  should not change nf.)  the other parameters are as
!                  described above and below.  calcf should not change
!                  n, p, or x.
! calcg.... (input) a subroutine that, given x, computes g(x), the gra-
!                  dient of f at x.  calcg must be declared external in
!                  the calling program.  it is invoked by
!                       call calcg(n, x, nf, g, uiparm, urparm, ufparm)
!                  when calcg is called, nf is the invocation
!                  count for calcf at the time f(x) was evaluated.  the
!                  x passed to calcg is usually the one passed to calcf
!                  on either its most recent invocation or the one
!                  prior to it.  if calcf saves intermediate results
!                  for use by calcg, then it is possible to tell from
!                  nf whether they are valid for the current x (or
!                  which copy is valid if two copies are kept).  if g
!                  cannot be computed at x, then calcg should set nf to
!                  0.  in this case, sumsl will return with iv(1) = 65.
!                  (if g can be computed at x, then calcg should not
!                  change nf.)  the other parameters to calcg are as
!                  described above and below.  calcg should not change
!                  n or x.
! iv....... (input/output) an integer value array of length liv (see
!                  below) that helps control the sumsl algorithm and
!                  that is used to store various intermediate quanti-
!                  ties.  of particular interest are the initialization/
!                  return code iv(1) and the entries in iv that control
!                  printing and limit the number of iterations and func-
!                  tion evaluations.  see the section on iv input
!                  values below.
! liv...... (input) length of iv array.  must be at least 60.  if liv
!                  is too small, then sumsl returns with iv(1) = 15.
!                  when sumsl returns, the smallest allowed value of
!                  liv is stored in iv(lastiv) -- see the section on
!                  iv output values below.  (this is intended for use
!                  with extensions of sumsl that handle constraints.)
! lv....... (input) length of v array.  must be at least 71+n*(n+15)/2.
!                  (at least 77+n*(n+17)/2 for smsno, at least
!                  78+n*(n+12) for humsl).  if lv is too small, then
!                  sumsl returns with iv(1) = 16.  when sumsl returns,
!                  the smallest allowed value of lv is stored in
!                  iv(lastv) -- see the section on iv output values
!                  below.
! v........ (input/output) a floating-point value array of length lv
!                  (see below) that helps control the sumsl algorithm
!                  and that is used to store various intermediate
!                  quantities.  of particular interest are the entries
!                  in v that limit the length of the first step
!                  attempted (lmax0) and specify convergence tolerances
!                  (afctol, lmaxs, rfctol, sctol, xctol, xftol).
! uiparm... (input) user integer parameter array passed without change
!                  to calcf and calcg.
! urparm... (input) user floating-point parameter array passed without
!                  change to calcf and calcg.
! ufparm... (input) user external subroutine or function passed without
!                  change to calcf and calcg.

!  ***  iv input values (from subroutine deflt)  ***

! iv(1)...  on input, iv(1) should have a value between 0 and 14......
!             0 and 12 mean this is a fresh start.  0 means that
!                  deflt(2, iv, liv, lv, v)
!             is to be called to provide all default values to iv and
!             v.  12 (the value that deflt assigns to iv(1)) means the
!             caller has already called deflt and has possibly changed
!             some iv and/or v entries to non-default values.
!             13 means deflt has been called and that sumsl (and
!             sumit) should only do their storage allocation.  that is,
!             they should set the output components of iv that tell
!             where various subarrays arrays of v begin, such as iv(g)
!             (and, for humsl and humit only, iv(dtol)), and return.
!             14 means that a storage has been allocated (by a call
!             with iv(1) = 13) and that the algorithm should be
!             started.  when called with iv(1) = 13, sumsl returns
!             iv(1) = 14 unless liv or lv is too small (or n is not
!             positive).  default = 12.
! iv(inith).... iv(25) tells whether the hessian approximation h should
!             be initialized.  1 (the default) means sumit should
!             initialize h to the diagonal matrix whose i-th diagonal
!             element is d(i)**2.  0 means the caller has supplied a
!             cholesky factor  l  of the initial hessian approximation
!             h = l*(l**t)  in v, starting at v(iv(lmat)) = v(iv(42))
!             (and stored compactly by rows).  note that iv(lmat) may
!             be initialized by calling sumsl with iv(1) = 13 (see
!             the iv(1) discussion above).  default = 1.
! iv(mxfcal)... iv(17) gives the maximum number of function evaluations
!             (calls on calcf) allowed.  if this number does not suf-
!             fice, then sumsl returns with iv(1) = 9.  default = 200.
! iv(mxiter)... iv(18) gives the maximum number of iterations allowed.
!             it also indirectly limits the number of gradient evalua-
!             tions (calls on calcg) to iv(mxiter) + 1.  if iv(mxiter)
!             iterations do not suffice, then sumsl returns with
!             iv(1) = 10.  default = 150.
! iv(outlev)... iv(19) controls the number and length of iteration sum-
!             mary lines printed (by itsum).  iv(outlev) = 0 means do
!             not print any summary lines.  otherwise, print a summary
!             line after each abs(iv(outlev)) iterations.  if iv(outlev)
!             is positive, then summary lines of length 78 (plus carri-
!             age control) are printed, including the following...  the
!             iteration and function evaluation counts, f = the current
!             function value, relative difference in function values
!             achieved by the latest step (i.e., reldf = (f0-v(f))/f01,
!             where f01 is the maximum of abs(v(f)) and abs(v(f0)) and
!             v(f0) is the function value from the previous itera-
!             tion), the relative function reduction predicted for the
!             step just taken (i.e., preldf = v(preduc) / f01, where
!             v(preduc) is described below), the scaled relative change
!             in x (see v(reldx) below), the step parameter for the
!             step just taken (stppar = 0 means a full newton step,
!             between 0 and 1 means a relaxed newton step, between 1
!             and 2 means a double dogleg step, greater than 2 means
!             a scaled down cauchy step -- see subroutine dbldog), the
!             2-norm of the scale vector d times the step just taken
!             (see v(dstnrm) below), and npreldf, i.e.,
!             v(nreduc)/f01, where v(nreduc) is described below -- if
!             npreldf is positive, then it is the relative function
!             reduction predicted for a newton step (one with
!             stppar = 0).  if npreldf is negative, then it is the
!             negative of the relative function reduction predicted
!             for a step computed with step bound v(lmaxs) for use in
!             testing for singular convergence.
!                  if iv(outlev) is negative, then lines of length 50
!             are printed, including only the first 6 items listed
!             above (through reldx).
!             default = 1.
! iv(parprt)... iv(20) = 1 means print any nondefault v values on a
!             fresh start or any changed v values on a restart.
!             iv(parprt) = 0 means skip this printing.  default = 1.
! iv(prunit)... iv(21) is the output unit number on which all printing
!             is done.  iv(prunit) = 0 means suppress all printing.
!             default = standard output unit (unit 6 on most systems).
! iv(solprt)... iv(22) = 1 means print out the value of x returned (as
!             well as the gradient and the scale vector d).
!             iv(solprt) = 0 means skip this printing.  default = 1.
! iv(statpr)... iv(23) = 1 means print summary statistics upon return-
!             ing.  these consist of the function value, the scaled
!             relative change in x caused by the most recent step (see
!             v(reldx) below), the number of function and gradient
!             evaluations (calls on calcf and calcg), and the relative
!             function reductions predicted for the last step taken and
!             for a newton step (or perhaps a step bounded by v(lmaxs)
!             -- see the descriptions of preldf and npreldf under
!             iv(outlev) above).
!             iv(statpr) = 0 means skip this printing.
!             iv(statpr) = -1 means skip this printing as well as that
!             of the one-line termination reason message.  default = 1.
! iv(x0prt).... iv(24) = 1 means print the initial x and scale vector d
!             (on a fresh start only).  iv(x0prt) = 0 means skip this
!             printing.  default = 1.

!  ***  (selected) iv output values  ***

! iv(1)........ on output, iv(1) is a return code....
!             3 = x-convergence.  the scaled relative difference (see
!                  v(reldx)) between the current parameter vector x and
!                  a locally optimal parameter vector is very likely at
!                  most v(xctol).
!             4 = relative function convergence.  the relative differ-
!                  ence between the current function value and its lo-
!                  cally optimal value is very likely at most v(rfctol).
!             5 = both x- and relative function convergence (i.e., the
!                  conditions for iv(1) = 3 and iv(1) = 4 both hold).
!             6 = absolute function convergence.  the current function
!                  value is at most v(afctol) in absolute value.
!             7 = singular convergence.  the hessian near the current
!                  iterate appears to be singular or nearly so, and a
!                  step of length at most v(lmaxs) is unlikely to yield
!                  a relative function decrease of more than v(sctol).
!             8 = false convergence.  the iterates appear to be converg-
!                  ing to a noncritical point.  this may mean that the
!                  convergence tolerances (v(afctol), v(rfctol),
!                  v(xctol)) are too small for the accuracy to which
!                  the function and gradient are being computed, that
!                  there is an error in computing the gradient, or that
!                  the function or gradient is discontinuous near x.
!             9 = function evaluation limit reached without other con-
!                  vergence (see iv(mxfcal)).
!            10 = iteration limit reached without other convergence
!                  (see iv(mxiter)).
!            11 = stopx returned .true. (external interrupt).  see the
!                  usage notes below.
!            14 = storage has been allocated (after a call with
!                  iv(1) = 13).
!            17 = restart attempted with n changed.
!            18 = d has a negative component and iv(dtype) .le. 0.
!            19...43 = v(iv(1)) is out of range.
!            63 = f(x) cannot be computed at the initial x.
!            64 = bad parameters passed to assess (which should not
!                  occur).
!            65 = the gradient could not be computed at x (see calcg
!                  above).
!            67 = bad first parameter to deflt.
!            80 = iv(1) was out of range.
!            81 = n is not positive.
! iv(g)........ iv(28) is the starting subscript in v of the current
!             gradient vector (the one corresponding to x).
! iv(lastiv)... iv(44) is the least acceptable value of liv.  (it is
!             only set if liv is at least 44.)
! iv(lastv).... iv(45) is the least acceptable value of lv.  (it is
!             only set if liv is large enough, at least iv(lastiv).)
! iv(nfcall)... iv(6) is the number of calls so far made on calcf (i.e.,
!             function evaluations).
! iv(ngcall)... iv(30) is the number of gradient evaluations (calls on
!             calcg).
! iv(niter).... iv(31) is the number of iterations performed.

!  ***  (selected) v input values (from subroutine deflt)  ***

! v(bias)..... v(43) is the bias parameter used in subroutine dbldog --
!             see that subroutine for details.  default = 0.8.
! v(afctol)... v(31) is the absolute function convergence tolerance.
!             if sumsl finds a point where the function value is less
!             than v(afctol) in absolute value, and if sumsl does not
!             return with iv(1) = 3, 4, or 5, then it returns with
!             iv(1) = 6.  this test can be turned off by setting
!             v(afctol) to zero.  default = max(10**-20, machep**2),
!             where machep is the unit roundoff.
! v(dinit).... v(38), if nonnegative, is the value to which the scale
!             vector d is initialized.  default = -1.
! v(lmax0).... v(35) gives the maximum 2-norm allowed for d times the
!             very first step that sumsl attempts.  this parameter can
!             markedly affect the performance of sumsl.
! v(lmaxs).... v(36) is used in testing for singular convergence -- if
!             the function reduction predicted for a step of length
!             bounded by v(lmaxs) is at most v(sctol) * abs(f0), where
!             f0  is the function value at the start of the current
!             iteration, and if sumsl does not return with iv(1) = 3,
!             4, 5, or 6, then it returns with iv(1) = 7.  default = 1.
! v(rfctol)... v(32) is the relative function convergence tolerance.
!             if the current model predicts a maximum possible function
!             reduction (see v(nreduc)) of at most v(rfctol)*abs(f0)
!             at the start of the current iteration, where  f0  is the
!             then current function value, and if the last step attempt-
!             ed achieved no more than twice the predicted function
!             decrease, then sumsl returns with iv(1) = 4 (or 5).
!             default = max(10**-10, machep**(2/3)), where machep is
!             the unit roundoff.
! v(sctol).... v(37) is the singular convergence tolerance -- see the
!             description of v(lmaxs) above.
! v(tuner1)... v(26) helps decide when to check for false convergence.
!             this is done if the actual function decrease from the
!             current step is no more than v(tuner1) times its predict-
!             ed value.  default = 0.1.
! v(xctol).... v(33) is the x-convergence tolerance.  if a newton step
!             (see v(nreduc)) is tried that has v(reldx) .le. v(xctol)
!             and if this step yields at most twice the predicted func-
!             tion decrease, then sumsl returns with iv(1) = 3 (or 5).
!             (see the description of v(reldx) below.)
!             default = machep**0.5, where machep is the unit roundoff.
! v(xftol).... v(34) is the false convergence tolerance.  if a step is
!             tried that gives no more than v(tuner1) times the predict-
!             ed function decrease and that has v(reldx) .le. v(xftol),
!             and if sumsl does not return with iv(1) = 3, 4, 5, 6, or
!             7, then it returns with iv(1) = 8.  (see the description
!             of v(reldx) below.)  default = 100*machep, where
!             machep is the unit roundoff.
! v(*)........ deflt supplies to v a number of tuning constants, with
!             which it should ordinarily be unnecessary to tinker.  see
!             section 17 of version 2.2 of the nl2sol usage summary
!             (i.e., the appendix to ref. 1) for details on v(i),
!             i = decfac, incfac, phmnfc, phmxfc, rdfcmn, rdfcmx,
!             tuner2, tuner3, tuner4, tuner5.

!  ***  (selected) v output values  ***

! v(dgnorm)... v(1) is the 2-norm of (diag(d)**-1)*g, where g is the
!             most recently computed gradient.
! v(dstnrm)... v(2) is the 2-norm of diag(d)*step, where step is the
!             current step.
! v(f)........ v(10) is the current function value.
! v(f0)....... v(13) is the function value at the start of the current
!             iteration.
! v(nreduc)... v(6), if positive, is the maximum function reduction
!             possible according to the current model, i.e., the func-
!             tion reduction predicted for a newton step (i.e.,
!             step = -h**-1 * g,  where  g  is the current gradient and
!             h is the current hessian approximation).
!                  if v(nreduc) is negative, then it is the negative of
!             the function reduction predicted for a step computed with
!             a step bound of v(lmaxs) for use in testing for singular
!             convergence.
! v(preduc)... v(7) is the function reduction predicted (by the current
!             quadratic model) for the current step.  this (divided by
!             v(f0)) is used in testing for relative function
!             convergence.
! v(reldx).... v(17) is the scaled relative change in x caused by the
!             current step, computed as
!                  max(abs(d(i)*(x(i)-x0(i)), 1 .le. i .le. p) /
!                     max(d(i)*(abs(x(i))+abs(x0(i))), 1 .le. i .le. p),
!             where x = x0 + step.

!-------------------------------  notes  -------------------------------

!  ***  algorithm notes  ***

!        this routine uses a hessian approximation computed from the
!     bfgs update (see ref 3).  only a cholesky factor of the hessian
!     approximation is stored, and this is updated using ideas from
!     ref. 4.  steps are computed by the double dogleg scheme described
!     in ref. 2.  the steps are assessed as in ref. 1.

!  ***  usage notes  ***

!        after a return with iv(1) .le. 11, it is possible to restart,
!     i.e., to change some of the iv and v input values described above
!     and continue the algorithm from the point where it was interrupt-
!     ed.  iv(1) should not be changed, nor should any entries of iv
!     and v other than the input values (those supplied by deflt).
!        those who do not wish to write a calcg which computes the
!     gradient analytically should call smsno rather than sumsl.
!     smsno uses finite differences to compute an approximate gradient.
!        those who would prefer to provide f and g (the function and
!     gradient) by reverse communication rather than by writing subrou-
!     tines calcf and calcg may call on sumit directly.  see the com-
!     ments at the beginning of sumit.
!        those who use sumsl interactively may wish to supply their
!     own stopx function, which should return .true. if the break key
!     has been pressed since stopx was last invoked.  this makes it
!     possible to externally interrupt sumsl (which will return with
!     iv(1) = 11 if stopx returns .true.).
!        storage for g is allocated at the end of v.  thus the caller
!     may make v longer than specified above and may allow calcg to use
!     elements of g beyond the first n as scratch storage.

!  ***  portability notes  ***

!      The sumsl distribution tape contains both single- and double-
!   precision versions of the sumsl source code, so it should be un-
!   necessary to change precisions.
!      Only the functions imdcon and rmdcon contain machine-dependent
!   constants.  to change from one machine to another, it should
!   suffice to change the (few) relevant lines in these functions.
!      The sumsl source code is expressed in 1966 ansi standard
!   fortran.  it may be converted to fortran 77 by commenting out all
!   lines that fall between a line having c/6 in columns 1-3 and a
!   line having c/7 in columns 1-3 and by removing (i.e., replacing
!   by a blank) the c in column 1 of the lines that follow the c/7
!   line and precede a line having c/ in columns 1-2 and blanks in
!   columns 3-72.  these changes convert some DATA statements into
!   parameter statements, convert some variables from real to
!   character*4, and make the DATA statements that initialize these
!   variables use character strings delimited by primes instead
!   of hollerith constants.  (such variables and DATA statements
!   appear only in modules itsum and parck.  parameter statements
!   appear nearly everywhere.)  these changes also add save statements
!   for variables given machine-dependent constants by rmdcon.

!  ***  references  ***

! 1.  dennis, j.e., gay, d.m., and welsch, r.e. (1981), algorithm 573 --
!             an adaptive nonlinear least-squares algorithm, acm trans.
!             math. software 7, pp. 369-383.

! 2.  dennis, j.e., and mei, h.h.w. (1979), two new unconstrained opti-
!             mization algorithms which use function and gradient
!             values, j. optim. theory applic. 28, pp. 453-482.

! 3.  dennis, j.e., and more, j.j. (1977), quasi-newton methods, motiva-
!             tion and theory, siam rev. 19, pp. 46-89.

! 4.  goldfarb, d. (1976), factorized variable metric methods for uncon-
!             strained optimization, math. comput. 30, pp. 796-811.

!  ***  general  ***

!     coded by david m. gay (winter 1980).  revised summer 1982.
!     this subroutine was written in connection with research
!     supported in part by the national science foundation under
!     grants mcs-7600324, dcr75-10143, 76-14311dss, mcs76-11989,
!     and mcs-7906671.
!.

!----------------------------  declarations  ---------------------------

INTEGER   :: g1,iv1,nf
REAL (dp) :: f

!  ***  subscripts for iv   ***

INTEGER, parameter :: nextv=47, nfcall=6, nfgcal=7, g=28, toobig=2, vneed=4

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

IF (iv(1) == 0) CALL deflt (2,iv,liv,lv,v)
iv1=iv(1)
IF (iv1 == 12 .OR. iv1 == 13) iv(vneed)=iv(vneed) + n
IF (iv1 == 14) GO TO 10
IF (iv1 > 2 .AND. iv1 < 12) GO TO 10
g1=1
IF (iv1 == 12) iv(1)=13
GO TO 20

10 g1=iv(g)

20 CALL sumit (d,f,v(g1:),iv,liv,lv,n,v,x)
IF (iv(1)-2 < 0) THEN
  GO TO 30
ELSE IF (iv(1)-2 == 0) THEN
  GO TO 40
ELSE
  GO TO 50
END IF

30 nf=iv(nfcall)
CALL calcf (n,x,nf,f,uiparm,urparm,ufparm)
IF (nf <= 0) iv(toobig)=1
GO TO 20

40 CALL calcg (n,x,iv(nfgcal),v(g1:),uiparm,urparm,ufparm)
GO TO 20

50 IF (iv(1) /= 14) GO TO 60

!  ***  storage allocation

iv(g)=iv(nextv)
iv(nextv)=iv(g) + n
IF (iv1 /= 13) GO TO 10

60 RETURN
!  ***  last card of sumsl follows  ***
END SUBROUTINE sumsl


SUBROUTINE smsno (n,d,x,calcf,iv,liv,lv,v,uiparm,urparm,ufparm)

!  ***  minimize general unconstrained objective function using
!  ***  finite-difference gradients and secant hessian approximations.

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: n,liv,lv
INTEGER, INTENT(IN OUT)   :: iv(:),uiparm(:)
REAL (dp), INTENT(IN OUT) :: d(:),x(:),v(:),urparm(:),ufparm

!     dimension v(77 + n*(n+17)/2), uiparm(*), urparm(*)

INTERFACE
  SUBROUTINE calcf(n, x, nf, f, uiparm, urparm, ufparm)
    USE machine_constants
    IMPLICIT NONE
    INTEGER, INTENT(IN)       :: n
    INTEGER, INTENT(IN OUT)   :: nf, uiparm(:)
    REAL (dp), INTENT(IN)     :: x(:)
    REAL (dp), INTENT(OUT)    :: f, urparm(:), ufparm
  END SUBROUTINE calcf
END INTERFACE

!  ***  purpose  ***

!        this routine interacts with subroutine  snoit  in an attempt
!     to find an n-vector  x*  that minimizes the (unconstrained)
!     objective function computed by  calcf.  (often the  x*  found is
!     a local minimizer rather than a global one.)

!  ***  parameters  ***

!        the parameters for smsno are the same as those for sumsl
!     (which see), except that calcg is omitted.  instead of calling
!     calcg to obtain the gradient of the objective function at x,
!     smsno calls sgrad2, which computes an approximation to the
!     gradient by finite (forward and central) differences using the
!     method of ref. 1.  the following input component is of interest
!     in this regard (and is not described in sumsl).

! v(eta0)..... v(42) is an estimated bound on the relative error in the
!             objective function value computed by calcf...
!                  (true value) = (computed value) * (1 + e),
!             where abs(e) .le. v(eta0).  default = machep * 10**3,
!             where machep is the unit roundoff.

!        the output values iv(nfcall) and iv(ngcall) have different
!     meanings for smsno than for sumsl...

! iv(nfcall)... iv(6) is the number of calls so far made on calcf (i.e.,
!             function evaluations) excluding those made only for
!             computing gradients.  the input value iv(mxfcal) is a
!             limit on iv(nfcall).
! iv(ngcall)... iv(30) is the number of function evaluations made only
!             for computing gradients.  the total number of function
!             evaluations is thus  iv(nfcall) + iv(ngcall).

!  ***  reference  ***

! 1. stewart, g.w. (1967), a modification of davidon*s minimization
!        method to accept difference approximations of derivatives,
!        j. assoc. comput. mach. 14, pp. 72-83.
!.
!  ***  general  ***

!     coded by david m. gay (winter 1980).  revised sept. 1982.
!     this subroutine was written in connection with research
!     supported in part by the national science foundation under
!     grants mcs-7600324, dcr75-10143, 76-14311dss, mcs76-11989,
!     and mcs-7906671.


!----------------------------  declarations  ---------------------------

INTEGER   :: nf
REAL (dp) :: fx

!  ***  subscripts for iv   ***

INTEGER, parameter :: nfcall=6, toobig=2

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

10 CALL snoit (d,fx,iv,liv,lv,n,v,x)
IF (iv(1) > 2) GO TO 20

!     ***  compute function  ***

nf=iv(nfcall)
CALL calcf (n,x,nf,fx,uiparm,urparm,ufparm)
IF (nf <= 0) iv(toobig)=1
GO TO 10


20 RETURN
!  ***  last card of smsno follows  ***
END SUBROUTINE smsno


SUBROUTINE sumit (d,fx,g,iv,liv,lv,n,v,x)

!  ***  carry out sumsl (unconstrained minimization) iterations, using
!  ***  double-dogleg/bfgs steps.

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: liv,lv,n
INTEGER, INTENT(IN OUT)   :: iv(:)
REAL (dp), INTENT(IN OUT) :: d(:),x(:),v(:),fx,g(:)

!--------------------------  parameter usage  --------------------------

! d.... scale vector.
! fx... function value.
! g.... gradient vector.
! iv... integer value array.
! liv.. length of iv (at least 60).
! lv... length of v (at least 71 + n*(n+13)/2).
! n.... number of variables (components in x and g).
! v.... floating-point value array.
! x.... vector of parameters to be optimized.

!  ***  discussion  ***

!        parameters iv, n, v, and x are the same as the corresponding
!     ones to sumsl (which see), except that v can be shorter (since
!     the part of v that sumsl uses for storing g is not needed).
!     moreover, compared with sumsl, iv(1) may have the two additional
!     output values 1 and 2, which are explained below, as is the use
!     of iv(toobig) and iv(nfgcal).  the value iv(g), which is an
!     output value from sumsl (and smsno), is not referenced by
!     sumit or the subroutines it calls.
!        fx and g need not have been initialized when sumit is called
!     with iv(1) = 12, 13, or 14.

! iv(1) = 1 means the caller should set fx to f(x), the function value
!             at x, and call sumit again, having changed none of the
!             other parameters.  an exception occurs if f(x) cannot be
!             (e.g. if overflow would occur), which may happen because
!             of an oversized step.  in this case the caller should set
!             iv(toobig) = iv(2) to 1, which will cause sumit to ig-
!             nore fx and try a smaller step.  the parameter nf that
!             sumsl passes to calcf (for possible use by calcg) is a
!             copy of iv(nfcall) = iv(6).
! iv(1) = 2 means the caller should set g to g(x), the gradient vector
!             of f at x, and call sumit again, having changed none of
!             the other parameters except possibly the scale vector d
!             when iv(dtype) = 0.  the parameter nf that sumsl passes
!             to calcg is iv(nfgcal) = iv(7).  if g(x) cannot be
!             evaluated, then the caller may set iv(nfgcal) to 0, in
!             which case sumit will return with iv(1) = 65.
!.
!  ***  general  ***

!     coded by david m. gay (december 1979).  revised sept. 1982.
!     this subroutine was written in connection with research supported
!     in part by the national science foundation under grants
!     mcs-7600324 and mcs-7906671.

!        (see sumsl for references.)

!+++++++++++++++++++++++++++  declarations  ++++++++++++++++++++++++++++

!  ***  local variables  ***

INTEGER   :: dg1,g01,i,k,l,lstgst,nwtst1,step1,temp1,w,x01,z
REAL (dp) :: t

!  ***  external functions and subroutines  ***

! assst.... assesses candidate step.
! dbdog.... computes double-dogleg (candidate) step.
! deflt.... supplies default iv and v input components.
! dotprd... returns inner product of two vectors.
! itsum.... prints iteration summary and info on initial and final x.
! litvmu... multiplies inverse transpose of lower triangle times vector.
! livmul... multiplies inverse of lower triangle times vector.
! ltvmul... multiplies transpose of lower triangle times vector.
! lupdt.... updates cholesky factor of hessian approximation.
! lvmul.... multiplies lower triangle times vector.
! parck.... checks validity of input iv and v values.
! reldst... computes v(reldx) = relative step size.
! stopx.... returns .true. if the break key has been pressed.
! vaxpy.... computes scalar times one vector plus another.
! vcopy.... copies one vector to another.
! vscopy... sets all elements of a vector to a scalar.
! vvmulp... multiplies vector by vector raised to power (componentwise).
! v2norm... returns the 2-norm of a vector.
! wzbfgs... computes w and z for lupdat corresponding to bfgs update.

!  ***  subscripts for iv and v  ***
!  ***  iv subscript values  ***

INTEGER :: cnvcod=55, dg=37, g0=48, inith=25, irc=29, kagqt=33, mode=35,   &
           model=5, mxfcal=17, mxiter=18, nfcall=6, nfgcal=7, ngcall=30,   &
           niter=31, nwtstp=34, radinc=8, restor=9, step=40, stglim=11,   &
           stlstg=41, toobig=2, vneed=4, xirc=13, x0=43

!  ***  v subscript values  ***

INTEGER :: dgnorm=1, dinit=38, dstnrm=2, dst0=3, f=10, f0=13, fdif=11, &
           gthg=44, gtstep=4, incfac=23, lmat=42, lmax0=35, lmaxs=36,  &
           nextv=47, nreduc=6, preduc=7, radfac=16, radius=8, rad0=9,  &
           reldx=17, tuner4=29, tuner5=30

REAL (dp) :: half=0.5_dp, negone=-1.d+0, one=1.d+0, onep2=1.2D+0, zero=0.d+0

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

i=iv(1)
IF (i == 1) GO TO 50
IF (i == 2) GO TO 60

!  ***  check validity of iv and v input values  ***

IF (iv(1) == 0) CALL deflt (2,iv,liv,lv,v)
IF (iv(1) == 12 .OR. iv(1) == 13) iv(vneed)=iv(vneed) + n*(n+13)/2
CALL parck (2,d,iv,liv,lv,n,v)
i=iv(1)-2
SELECT CASE (i)
  CASE (1:6)
    GO TO 180
  CASE (7)
    GO TO 120
  CASE (8)
    GO TO 90
  CASE (9)
    GO TO 120
  CASE (10:11)
    GO TO 10
  CASE (12)
    GO TO 20
  CASE DEFAULT
    GO TO 310
END SELECT

!  ***  storage allocation  ***

10 l=iv(lmat)
iv(x0)=l + n*(n+1)/2
iv(step)=iv(x0) + n
iv(stlstg)=iv(step) + n
iv(g0)=iv(stlstg) + n
iv(nwtstp)=iv(g0) + n
iv(dg)=iv(nwtstp) + n
iv(nextv)=iv(dg) + n
IF (iv(1) /= 13) GO TO 20
iv(1)=14
GO TO 310

!  ***  initialization  ***

20 iv(niter)=0
iv(nfcall)=1
iv(ngcall)=1
iv(nfgcal)=1
iv(mode)=-1
iv(model)=1
iv(stglim)=1
iv(toobig)=0
iv(cnvcod)=0
iv(radinc)=0
v(rad0)=zero
IF (v(dinit) >= zero) CALL vscopy (n,d,v(dinit))
IF (iv(inith) /= 1) GO TO 40

!     ***  set the initial hessian approximation to diag(d)**-2  ***

l=iv(lmat)
CALL vscopy (n*(n+1)/2,v(l:),zero)
k=l-1
DO i=1,n
  k=k + i
  t=d(i)
  IF (t <= zero) t=one
  v(k)=t
END DO

!  ***  compute initial function value  ***

40 iv(1)=1
GO TO 310

50 v(f)=fx
IF (iv(mode) >= 0) GO TO 180
iv(1)=2
IF (iv(toobig) == 0) GO TO 310
iv(1)=63
GO TO 300

!  ***  make sure gradient could be computed  ***

60 IF (iv(nfgcal) /= 0) GO TO 70
iv(1)=65
GO TO 300

70 dg1=iv(dg)
CALL vvmulp (n,v(dg1:),g,d,-1)
v(dgnorm)=v2norm(n,v(dg1:))

IF (iv(cnvcod) /= 0) GO TO 290
IF (iv(mode) == 0) GO TO 250

!  ***  allow first step to have scaled 2-norm at most v(lmax0)  ***

v(radius)=v(lmax0)

iv(mode)=0


!-----------------------------  main loop  -----------------------------


!  ***  print iteration summary, check iteration limit  ***

80 CALL itsum (d,g,iv,n,v,x)
90 k=iv(niter)
IF (k < iv(mxiter)) GO TO 100
iv(1)=10
GO TO 300

!  ***  update radius  ***

100 iv(niter)=k + 1
IF (k > 0) v(radius)=v(radfac)*v(dstnrm)

!  ***  initialize for start of next iteration  ***

g01=iv(g0)
x01=iv(x0)
v(f0)=v(f)
iv(irc)=4
iv(kagqt)=-1

!     ***  copy x to x0, g to g0 ***

CALL vcopy (n,v(x01:),x)
CALL vcopy (n,v(g01:),g)

!  ***  check stopx and function evaluation limit  ***

110 IF (.NOT.stopx()) GO TO 130
iv(1)=11
GO TO 140

!     ***  come here when restarting after func. eval. limit or stopx.

120 IF (v(f) >= v(f0)) GO TO 130
v(radfac)=one
k=iv(niter)
GO TO 100

130 IF (iv(nfcall) < iv(mxfcal)) GO TO 150
iv(1)=9
140 IF (v(f) >= v(f0)) GO TO 300

!        ***  in case of stopx or function evaluation limit with
!        ***  improved v(f), evaluate the gradient at x.

iv(cnvcod)=iv(1)
GO TO 240

!. . . . . . . . . . . . .  compute candidate step  . . . . . . . . . .

150 step1=iv(step)
dg1=iv(dg)
nwtst1=iv(nwtstp)
IF (iv(kagqt) >= 0) GO TO 160
l=iv(lmat)
CALL livmul (n,v(nwtst1:),v(l:),g)
v(nreduc)=half*dotprd(n,v(nwtst1:),v(nwtst1:))
CALL litvmu (n,v(nwtst1:),v(l:),v(nwtst1:))
CALL vvmulp (n,v(step1:),v(nwtst1:),d,1)
v(dst0)=v2norm(n,v(step1:))
CALL vvmulp (n,v(dg1:),v(dg1:),d,-1)
CALL ltvmul (n,v(step1:),v(l:),v(dg1:))
v(gthg)=v2norm(n,v(step1:))
iv(kagqt)=0
160 CALL dbdog (v(dg1:),n,v(nwtst1:),v(step1:),v)
IF (iv(irc) == 6) GO TO 180

!  ***  check whether evaluating f(x0 + step) looks worthwhile  ***

IF (v(dstnrm) <= zero) GO TO 180
IF (iv(irc) /= 5) GO TO 170
IF (v(radfac) <= one) GO TO 170
IF (v(preduc) <= onep2*v(fdif)) GO TO 180

!  ***  compute f(x0 + step)  ***

170 x01=iv(x0)
step1=iv(step)
CALL vaxpy (n,x,one,v(step1:),v(x01:))
iv(nfcall)=iv(nfcall) + 1
iv(1)=1
iv(toobig)=0
GO TO 310

!. . . . . . . . . . . . .  assess candidate step  . . . . . . . . . . .

180 x01=iv(x0)
v(reldx)=reldst(n,d,x,v(x01:))
CALL assst (iv,v)
step1=iv(step)
lstgst=iv(stlstg)
IF (iv(restor) == 1) CALL vcopy (n,x,v(x01:))
IF (iv(restor) == 2) CALL vcopy (n,v(lstgst:),v(step1:))
IF (iv(restor) /= 3) GO TO 190
CALL vcopy (n,v(step1:),v(lstgst:))
CALL vaxpy (n,x,one,v(step1:),v(x01:))
v(reldx)=reldst(n,d,x,v(x01:))

190 k=iv(irc)

SELECT CASE (k)
  CASE (2:4)
    GO TO 230
  CASE (6)
    GO TO 210
  CASE (7:12)
    GO TO 220
  CASE (13)
    GO TO 280
  CASE (14)
    GO TO 250
  CASE DEFAULT
    GO TO 200
END SELECT

!     ***  recompute step with changed radius  ***

200 v(radius)=v(radfac)*v(dstnrm)
GO TO 110

!  ***  compute step of length v(lmaxs) for singular convergence test.

210 v(radius)=v(lmaxs)
GO TO 150

!  ***  convergence or false convergence  ***

220 iv(cnvcod)=k-4
IF (v(f) >= v(f0)) GO TO 290
IF (iv(xirc) == 14) GO TO 290
iv(xirc)=14

!. . . . . . . . . . . .  process acceptable step  . . . . . . . . . . .

230 IF (iv(irc) /= 3) GO TO 240
step1=iv(step)
temp1=iv(stlstg)

!     ***  set  temp1 = hessian * step  for use in gradient tests  ***

l=iv(lmat)
CALL ltvmul (n,v(temp1:),v(l:),v(step1:))
CALL lvmul (n,v(temp1:),v(l:),v(temp1:))

!  ***  compute gradient  ***

240 iv(ngcall)=iv(ngcall) + 1
iv(1)=2
GO TO 310

!  ***  initializations -- g0 = g - g0, etc.  ***

250 g01=iv(g0)
CALL vaxpy (n,v(g01:),negone,v(g01:),g)
step1=iv(step)
temp1=iv(stlstg)
IF (iv(irc) /= 3) GO TO 270

!  ***  set v(radfac) by gradient tests  ***

!     ***  set  temp1 = diag(d)**-1 * (hessian*step + (g(x0)-g(x)))  ***

CALL vaxpy (n,v(temp1:),negone,v(g01:),v(temp1:))
CALL vvmulp (n,v(temp1:),v(temp1:),d,-1)

!        ***  do gradient tests  ***

IF (v2norm(n,v(temp1:)) <= v(dgnorm)*v(tuner4)) GO TO 260
IF (dotprd(n,g,v(step1:)) >= v(gtstep)*v(tuner5)) GO TO 270
260 v(radfac)=v(incfac)

!  ***  update h, loop  ***

270 w=iv(nwtstp)
z=iv(x0)
l=iv(lmat)
CALL wzbfgs (v(l:),n,v(step1:),v(w:),v(g01:),v(z:))

!     ** use the n-vectors starting at v(step1) and v(g01) for scratch..
CALL lupdat (v(temp1:),v(step1:),v(l:),v(g01:),v(l:),n,v(w:),v(z:))
iv(1)=2
GO TO 80

!. . . . . . . . . . . . . .  misc. details  . . . . . . . . . . . . . .

!  ***  bad parameters to assess  ***

280 iv(1)=64
GO TO 300

!  ***  print summary of final iteration and other requested items  ***

290 iv(1)=iv(cnvcod)
iv(cnvcod)=0
300 CALL itsum (d,g,iv,n,v,x)

310 RETURN

!  ***  last line of sumit follows  ***
END SUBROUTINE sumit


SUBROUTINE snoit (d,fx,iv,liv,lv,n,v,x)

!  ***  iteration driver for smsno...
!  ***  minimize general unconstrained objective function using
!  ***  finite-difference gradients and secant hessian approximations.

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: liv,lv,n
INTEGER, INTENT(IN OUT)   :: iv(:)
REAL (dp), INTENT(IN OUT) :: d(:),x(:),v(:),fx
!     dimension v(77 + n*(n+17)/2)

!  ***  purpose  ***

!        this routine interacts with subroutine  sumit  in an attempt
!     to find an n-vector  x*  that minimizes the (unconstrained)
!     objective function  fx = f(x)  computed by the caller.  (often
!     the  x*  found is a local minimizer rather than a global one.)

!  ***  parameters  ***

!        the parameters for snoit are the same as those for sumsl
!     (which see), except that calcf, calcg, uiparm, urparm, and ufparm
!     are omitted, and a parameter  fx  for the objective function
!     value at x is added.  instead of calling calcg to obtain the
!     gradient of the objective function at x, snoit calls sgrad2,
!     which computes an approximation to the gradient by finite
!     (forward and central) differences using the method of ref. 1.
!     the following input component is of interest in this regard
!     (and is not described in sumsl).

! v(eta0)..... v(42) is an estimated bound on the relative error in the
!             objective function value computed by calcf...
!                  (true value) = (computed value) * (1 + e),
!             where abs(e) .le. v(eta0).  default = machep * 10**3,
!             where machep is the unit roundoff.

!        the output values iv(nfcall) and iv(ngcall) have different
!     meanings for smsno than for sumsl...

! iv(nfcall)... iv(6) is the number of calls so far made on calcf (i.e.,
!             function evaluations) excluding those made only for
!             computing gradients.  the input value iv(mxfcal) is a
!             limit on iv(nfcall).
! iv(ngcall)... iv(30) is the number of function evaluations made only
!             for computing gradients.  the total number of function
!             evaluations is thus  iv(nfcall) + iv(ngcall).

!  ***  references  ***

! 1. stewart, g.w. (1967), a modification of davidon*s minimization
!        method to accept difference approximations of derivatives,
!        j. assoc. comput. mach. 14, pp. 72-83.
!.
!  ***  general  ***

!     coded by david m. gay (august 1982).

!----------------------------  declarations  ---------------------------

! deflt.... supplies default parameter values.
! dotprd... returns inner product of two vectors.
! sgrad2... computes finite-difference gradient approximation.
! sumit.... reverse-communication routine that does sumsl algorithm.
! vscopy... sets all elements of a vector to a scalar.

INTEGER   :: alpha,g1,i,iv1,j,k,w
REAL (dp) :: zero=0.d+0

!  ***  subscripts for iv   ***

INTEGER :: eta0=42, f=10, g=28, lmat=42, nextv=47, nfgcal=7, ngcall=30,  &
           niter=31, sgirc=57, toobig=2, vneed=4

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

iv1=iv(1)
IF (iv1 == 1) GO TO 10
IF (iv1 == 2) GO TO 50
IF (iv(1) == 0) CALL deflt (2,iv,liv,lv,v)
iv1=iv(1)
IF (iv1 == 12 .OR. iv1 == 13) iv(vneed)=iv(vneed) + 2*n + 6
IF (iv1 == 14) GO TO 10
IF (iv1 > 2 .AND. iv1 < 12) GO TO 10
g1=1
IF (iv1 == 12) iv(1)=13
GO TO 20

10 g1=iv(g)

20 CALL sumit (d,fx,v(g1:),iv,liv,lv,n,v,x)
IF (iv(1)-2 < 0) THEN
  GO TO 80
ELSE IF (iv(1)-2 == 0) THEN
  GO TO 30
ELSE
  GO TO 70
END IF

!  ***  compute gradient  ***

30 IF (iv(niter) == 0) CALL vscopy (n,v(g1:),zero)
j=iv(lmat)
k=g1-n
DO i=1,n
  v(k)=dotprd(i,v(j:),v(j:))
  k=k + 1
  j=j + i
END DO
!     ***  undo increment of iv(ngcall) done by sumit  ***
iv(ngcall)=iv(ngcall)-1
!     ***  store return code from sgrad2 in iv(sgirc)  ***
iv(sgirc)=0
!     ***  x may have been restored, so copy back fx... ***
fx=v(f)
GO TO 60

!     ***  gradient loop  ***

50 IF (iv(toobig) == 0) GO TO 60
iv(nfgcal)=0
GO TO 10

60 g1=iv(g)
alpha=g1-n
w=alpha-6
CALL sgrad2 (v(alpha:),d,v(eta0),fx,v(g1:),iv(sgirc),n,v(w:),x)
IF (iv(sgirc) == 0) GO TO 10
iv(ngcall)=iv(ngcall) + 1
GO TO 80

70 IF (iv(1) /= 14) GO TO 80

!  ***  storage allocation  ***

iv(g)=iv(nextv) + n + 6
iv(nextv)=iv(g) + n
IF (iv1 /= 13) GO TO 10

80 RETURN
!  ***  last card of snoit follows  ***
END SUBROUTINE snoit


SUBROUTINE dbdog (dig,n,nwtstp,step,v)

!  ***  compute double dogleg step  ***

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: n
REAL (dp), INTENT(IN)     :: dig(:),nwtstp(:)
REAL (dp), INTENT(IN OUT) :: v(:)
REAL (dp), INTENT(OUT)    :: step(:)

!  ***  purpose  ***

!      This subroutine computes a candidate step (for use in an unconstrained
!   minimization code) by the double dogleg algorithm of Dennis and Mei
!   (ref. 1), which is a variation on Powell's dogleg scheme (ref. 2, p. 95).

!--------------------------  parameter usage  --------------------------

!    dig (input) diag(d)**-2 * g -- see algorithm notes.
!      g (input) the current gradient vector.
!      n (input) number of components in  dig, g, nwtstp,  and  step.
! nwtstp (input) negative newton step -- see algorithm notes.
!   step (output) the computed step.
!      v (i/o) values array, the following components of which are
!             used here...
! v(bias)   (input) bias for relaxed newton step, which is v(bias) of
!             the way from the full newton to the fully relaxed newton
!             step.  recommended value = 0.8 .
! v(dgnorm) (input) 2-norm of diag(d)**-1 * g -- see algorithm notes.
! v(dstnrm) (output) 2-norm of diag(d) * step, which is v(radius)
!             unless v(stppar) = 0 -- see algorithm notes.
! v(dst0) (input) 2-norm of diag(d) * nwtstp -- see algorithm notes.
! v(grdfac) (output) the coefficient of  dig  in the step returned --
!             step(i) = v(grdfac)*dig(i) + v(nwtfac)*nwtstp(i).
! v(gthg)   (input) square-root of (dig**t) * (hessian) * dig -- see
!             algorithm notes.
! v(gtstep) (output) inner product between g and step.
! v(nreduc) (output) function reduction predicted for the full newton
!             step.
! v(nwtfac) (output) the coefficient of  nwtstp  in the step returned --
!             see v(grdfac) above.
! v(preduc) (output) function reduction predicted for the step returned.
! v(radius) (input) the trust region radius.  d times the step returned
!             has 2-norm v(radius) unless v(stppar) = 0.
! v(stppar) (output) code telling how step was computed... 0 means a
!             full newton step.  between 0 and 1 means v(stppar) of the
!             way from the newton to the relaxed newton step.  between
!             1 and 2 means a true double dogleg step, v(stppar) - 1 of
!             the way from the relaxed newton to the cauchy step.
!             greater than 2 means 1 / (v(stppar) - 1) times the cauchy
!             step.

!-------------------------------  notes  -------------------------------

!  ***  algorithm notes  ***

!      Let  g  and  h  be the current gradient and hessian approximation
!   respectively and let d be the current scale vector.  This routine
!   assumes dig = diag(d)**-2 * g  and  nwtstp = h**-1 * g.   The step
!   computed is the same one would get by replacing g and h with
!   diag(d)**-1 * g  and  diag(d)**-1 * h * diag(d)**-1, computing step,
!   and translating step back to the original variables,
!   i.e., premultiplying by diag(d)**-1.

!  ***  references  ***

! 1.  dennis, j.e., and mei, h.h.w. (1979), two new unconstrained opti-
!             mization algorithms which use function and gradient
!             values, j. optim. theory applic. 28, pp. 453-482.
! 2.  powell, m.j.d. (1970), a hybrid method for non-linear equations,
!             in numerical methods for non-linear equations, edited by
!             p. rabinowitz, gordon and breach, london.

!  ***  general  ***

!     coded by david m. gay.
!     this subroutine was written in connection with research supported
!     by the national science foundation under grants mcs-7600324 and
!     mcs-7906671.

!--------------------------  local variables  --------------------------

INTEGER   :: i
REAL (dp) :: cfact,cnorm,ctrnwt,ghinvg,femnsq,gnorm,nwtnrm,  &
             relax,rlambd,t,t1,t2
REAL (dp) :: half= 0.5_dp, one=1._dp, two=2._dp, zero=0._dp

!  ***  v subscripts  ***

INTEGER   :: bias=43, dgnorm=1, dstnrm=2, dst0=3, grdfac=45, gthg=44,  &
             gtstep=4, nreduc=6, nwtfac=46, preduc=7, radius=8, stppar=5

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

nwtnrm=v(dst0)
rlambd=one
IF (nwtnrm > zero) rlambd=v(radius)/nwtnrm
gnorm=v(dgnorm)
ghinvg=two*v(nreduc)
v(grdfac)=zero
v(nwtfac)=zero
IF (rlambd < one) GO TO 20

!        ***  the newton step is inside the trust region  ***

v(stppar)=zero
v(dstnrm)=nwtnrm
v(gtstep)=-ghinvg
v(preduc)=v(nreduc)
v(nwtfac)=-one
DO i=1,n
  step(i)=-nwtstp(i)
END DO
GO TO 80

20 v(dstnrm)=v(radius)
cfact=(gnorm/v(gthg))**2
!     ***  cauchy step = -cfact * g.
cnorm=gnorm*cfact
relax=one-v(bias)*(one-gnorm*cnorm/ghinvg)
IF (rlambd < relax) GO TO 40

!        ***  step is between relaxed newton and full newton steps  ***

v(stppar)=one-(rlambd-relax)/(one-relax)
t=-rlambd
v(gtstep)=t*ghinvg
v(preduc)=rlambd*(one-half*rlambd)*ghinvg
v(nwtfac)=t
DO i=1,n
  step(i)=t*nwtstp(i)
END DO
GO TO 80

40 IF (cnorm < v(radius)) GO TO 60

!        ***  the cauchy step lies outside the trust region --
!        ***  step = scaled cauchy step  ***

t=-v(radius)/gnorm
v(grdfac)=t
v(stppar)=one + cnorm/v(radius)
v(gtstep)=-v(radius)*gnorm
v(preduc)=v(radius)*(gnorm-half*v(radius)*(v(gthg)/gnorm)**2)
DO i=1,n
  step(i)=t*dig(i)
END DO
GO TO 80

!     ***  compute dogleg step between cauchy and relaxed newton  ***
!     ***  femur = relaxed newton step minus cauchy step  ***

60 ctrnwt=cfact*relax*ghinvg/gnorm
!     *** ctrnwt = inner prod. of cauchy and relaxed newton steps,
!     *** scaled by gnorm**-1.
t1=ctrnwt-gnorm*cfact**2
!     ***  t1 = inner prod. of femur and cauchy step, scaled by
!     ***  gnorm**-1.
t2=v(radius)*(v(radius)/gnorm)-gnorm*cfact**2
t=relax*nwtnrm
femnsq=(t/gnorm)*t-ctrnwt-t1
!     ***  femnsq = square of 2-norm of femur, scaled by gnorm**-1.
t=t2/(t1 + SQRT(t1**2 + femnsq*t2))
!     ***  dogleg step  =  cauchy step  +  t * femur.
t1=(t-one)*cfact
v(grdfac)=t1
t2=-t*relax
v(nwtfac)=t2
v(stppar)=two-t
v(gtstep)=t1*gnorm**2 + t2*ghinvg
v(preduc)=-t1*gnorm*((t2 + one)*gnorm) - t2*(one + half*t2)*ghinvg -   &
          half*(v(gthg)*t1)**2
step(1:n)=t1*dig(1:n) + t2*nwtstp(1:n)

80 RETURN
!  ***  last line of dbdog follows  ***
END SUBROUTINE dbdog


SUBROUTINE ltvmul (n,x,l,y)

!  ***  compute  x = (l**t)*y, where  l  is an  n x n  lower
!  ***  triangular matrix stored compactly by rows.  x and y may
!  ***  occupy the same storage.  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: n
REAL (dp), INTENT(IN)  :: l(:),y(:)
REAL (dp), INTENT(OUT) :: x(:)
!     dimension l(n*(n+1)/2)

! Local variables
INTEGER   :: i,ij,i0,j
REAL (dp) :: yi,zero=0.d+0

i0=0
DO i=1,n
  yi=y(i)
  x(i)=zero
  DO j=1,i
    ij=i0 + j
    x(j)=x(j) + yi*l(ij)
  END DO
  i0=i0 + i
END DO
RETURN
!  ***  last card of ltvmul follows  ***
END SUBROUTINE ltvmul


SUBROUTINE lupdat (beta,gamma,l,lambda,lplus,n,w,z)

!  ***  compute lplus = secant update of l  ***

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: n
REAL (dp), INTENT(IN OUT) :: beta(:),gamma(:),l(:),lambda(:),lplus(:),  &
                             w(:),z(:)
!     dimension l(n*(n+1)/2), lplus(n*(n+1)/2)

!--------------------------  parameter usage  --------------------------

!   beta = scratch vector.
!  gamma = scratch vector.
!      l (input) lower triangular matrix, stored rowwise.
! lambda = scratch vector.
!  lplus (output) lower triangular matrix, stored rowwise, which may
!             occupy the same storage as  l.
!      n (input) length of vector parameters and order of matrices.
!      w (input, destroyed on output) right singular vector of rank 1
!             correction to  l.
!      z (input, destroyed on output) left singular vector of rank 1
!             correction to  l.

!-------------------------------  notes  -------------------------------

!  ***  application and usage restrictions  ***

!        this routine updates the cholesky factor  l  of a symmetric
!     positive definite matrix to which a secant update is being
!     applied -- it computes a cholesky factor  lplus  of
!     l * (i + z*w**t) * (i + w*z**t) * l**t.  it is assumed that  w
!     and  z  have been chosen so that the updated matrix is strictly
!     positive definite.

!  ***  algorithm notes  ***

!        this code uses recurrence 3 of ref. 1 (with d(j) = 1 for all j)
!     to compute  lplus  of the form  l * (i + z*w**t) * q,  where  q
!     is an orthogonal matrix that makes the result lower triangular.
!        lplus may have some negative diagonal elements.

!  ***  references  ***

! 1.  goldfarb, d. (1976), factorized variable metric methods for uncon-
!             strained optimization, math. comput. 30, pp. 796-811.

!  ***  general  ***

!     coded by david m. gay (fall 1979).
!     this subroutine was written in connection with research supported
!     by the national science foundation under grants mcs-7600324 and
!     mcs-7906671.

!--------------------------  local variables  --------------------------

INTEGER   :: i,ij,j,jj,jp1,k,nm1,np1
REAL (dp) :: a,b,bj,eta,gj,lj,lij,ljj,nu,s,theta,wj,zj

REAL (dp), parameter :: one=1.d+0, zero=0.d+0

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

nu=one
eta=zero
IF (n <= 1) GO TO 30
nm1=n-1

!  ***  temporarily store s(j) = sum over k = j+1 to n of w(k)**2 in
!  ***  lambda(j).

s=zero
DO i=1,nm1
  j=n-i
  s=s + w(j+1)**2
  lambda(j)=s
END DO

!  ***  compute lambda, gamma, and beta by goldfarb*s recurrence 3.

DO j=1,nm1
  wj=w(j)
  a=nu*z(j)-eta*wj
  theta=one + a*wj
  s=a*lambda(j)
  lj=SQRT(theta**2 + a*s)
  IF (theta > zero) lj=-lj
  lambda(j)=lj
  b=theta*wj + s
  gamma(j)=b*nu/lj
  beta(j)=(a-b*eta)/lj
  nu=-nu/lj
  eta=-(eta + (a**2)/(theta-lj))/lj
END DO
30 lambda(n)=one + (nu*z(n)-eta*w(n))*w(n)

!  ***  update l, gradually overwriting  w  and  z  with  l*w  and  l*z.

np1=n + 1
jj=n*(n+1)/2
DO k=1,n
  j=np1-k
  lj=lambda(j)
  ljj=l(jj)
  lplus(jj)=lj*ljj
  wj=w(j)
  w(j)=ljj*wj
  zj=z(j)
  z(j)=ljj*zj
  IF (k == 1) GO TO 50
  bj=beta(j)
  gj=gamma(j)
  ij=jj + j
  jp1=j + 1
  DO i=jp1,n
    lij=l(ij)
    lplus(ij)=lj*lij + bj*w(i) + gj*z(i)
    w(i)=w(i) + lij*wj
    z(i)=z(i) + lij*zj
    ij=ij + i
  END DO
  50 jj=jj-j
END DO

RETURN
!  ***  last card of lupdat follows  ***
END SUBROUTINE lupdat


SUBROUTINE lvmul (n,x,l,y)

!  ***  compute  x = l*y, where  l  is an  n x n  lower triangular
!  ***  matrix stored compactly by rows.  x and y may occupy the same
!  ***  storage.  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: n
REAL (dp), INTENT(IN)  :: l(:),y(:)
REAL (dp), INTENT(OUT) :: x(:)
!     dimension l(n*(n+1)/2)

! Local variables
INTEGER   :: i,ii,ij,i0,j,np1
REAL (dp) :: t, zero= 0._dp

np1=n + 1
i0=n*(n+1)/2
DO ii=1,n
  i=np1-ii
  i0=i0-i
  t=zero
  DO j=1,i
    ij=i0 + j
    t=t + l(ij)*y(j)
  END DO
  x(i)=t
END DO
RETURN
!  ***  last card of lvmul follows  ***
END SUBROUTINE lvmul


SUBROUTINE sgrad2 (alpha,d,eta0,fx,g,irc,n,w,x)

!  ***  compute finite difference gradient by stewart*s scheme  ***

!     ***  parameters  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: n
INTEGER, INTENT(IN OUT)   :: irc
REAL (dp), INTENT(IN)     :: alpha(:),d(:),eta0
REAL (dp), INTENT(IN OUT) :: fx,g(:),w(:),x(:)

!.......................................................................

!     ***  purpose  ***

!        this subroutine uses an embellished form of the finite-differ-
!     ence scheme proposed by stewart (ref. 1) to approximate the
!     gradient of the function f(x), whose values are supplied by
!     reverse communication.

!     ***  parameter description  ***

!  alpha in  (approximate) diagonal elements of the hessian of f(x).
!      d in  scale vector such that d(i)*x(i), i = 1,...,n, are in
!             comparable units.
!   eta0 in  estimated bound on relative error in the function value...
!             (true value) = (computed value)*(1+e), where abs(e) <= eta0.
!     fx i/o on input,  fx  must be the computed value of f(x).  on
!             output with irc = 0, fx has been restored to its original
!             value, the one it had when sgrad2 was last called with
!             irc = 0.
!      g i/o on input with irc = 0, g should contain an approximation
!             to the gradient of f near x, e.g., the gradient at the
!             previous iterate.  when sgrad2 returns with irc = 0, g is
!             the desired finite-difference approximation to the
!             gradient at x.
!    irc i/o input/return code... before the very first call on sgrad2,
!             the caller must set irc to 0.  whenever sgrad2 returns a
!             nonzero value for irc, it has perturbed some component of
!             x... the caller should evaluate f(x) and call sgrad2
!             again with fx = f(x).
!      n in  the number of variables (components of x) on which f
!             depends.
!      x i/o on input with irc = 0, x is the point at which the
!             gradient of f is desired.  on output with irc nonzero, x
!             is the point at which f should be evaluated.  on output
!             with irc = 0, x has been restored to its original value
!             (the one it had when sgrad2 was last called with irc = 0)
!             and g contains the desired gradient approximation.
!      w i/o work vector of length 6 in which sgrad2 saves certain
!             quantities while the caller is evaluating f(x) at a
!             perturbed x.

!     ***  application and usage restrictions  ***

!        this routine is intended for use with quasi-newton routines
!     for unconstrained minimization (in which case  alpha  comes from
!     the diagonal of the quasi-newton hessian approximation).

!     ***  algorithm notes  ***

!        this code departs from the scheme proposed by stewart (ref. 1)
!     in its guarding against overly large or small step sizes and its
!     handling of special cases (such as zero components of alpha or g).

!     ***  references  ***

! 1. stewart, g.w. (1967), a modification of davidon*s minimization
!        method to accept difference approximations of derivatives,
!        j. assoc. comput. mach. 14, pp. 72-83.

!     ***  history  ***

!     designed and coded by david m. gay (summer 1977/summer 1980).

!     ***  general  ***

!        this routine was prepared in connection with work supported by
!     the national science foundation under grants mcs76-00324 and
!     mcs-7906671.

!.......................................................................

!     ***** local variables *****

REAL (dp) :: aai,afx,afxeta,agi,alphai,axi,axibar,discon,eta,gi,h,hmin

REAL (dp) :: c2000=2.0D+3, four=4.0D+0, hmax0=0.02D+0, hmin0=5.0D+1, h0,  &
             machep, one=1.0D+0, p002=0.002D+0, three=3.0D+0, two=2.0D+0, &
             zero=0.0D+0
INTEGER   :: i, fh=3, fx0=4, hsave=5, xisave=6

!---------------------------------  body  ------------------------------

IF (irc < 0) THEN
  GO TO 50
ELSE IF (irc == 0) THEN
  GO TO 10
ELSE
  GO TO 110
END IF

!     ***  fresh start -- get machine-dependent constants  ***

!   store machep in w(1) and h0 in w(2), where machep is the unit roundoff
!   (the smallest positive number such that 1 + machep > 1  and
!   1 - machep < 1),  and  h0 is the square-root of machep.

10 w(1)=rmdcon(3)
w(2)=SQRT(w(1))

w(fx0)=fx

!     ***  increment  i  and start computing  g(i)  ***

20 i=ABS(irc)+1
IF (i > n) GO TO 120
irc=i
afx=ABS(w(fx0))
machep=w(1)
h0=w(2)
hmin=hmin0*machep
w(xisave)=x(i)
axi=ABS(x(i))
axibar=MAX(axi,one/d(i))
gi=g(i)
agi=ABS(gi)
eta=ABS(eta0)
IF (afx > zero) eta=MAX(eta,agi*axi*machep/afx)
alphai=alpha(i)
IF (alphai == zero) GO TO 80
IF (gi == zero .OR. fx == zero) GO TO 90
afxeta=afx*eta
aai=ABS(alphai)

!        *** compute h = stewart*s forward-difference step size.

IF (gi**2 <= afxeta*aai) GO TO 30
h=two*SQRT(afxeta/aai)
h=h*(one-aai*h/(three*aai*h+four*agi))
GO TO 40
30 h=two*(afxeta*agi/(aai**2))**(one/three)
h=h*(one-two*agi/(three*aai*h+four*agi))

!        ***  ensure that  h  is not insignificantly small  ***

40 h=MAX(h,hmin*axibar)

!        *** use forward difference if bound on truncation error is at
!        *** most 10**-3.

IF (aai*h <= p002*agi) GO TO 70

!        *** compute h = stewart*s step for central difference.

discon=c2000*afxeta
h=discon/(agi+SQRT(gi**2+aai*discon))

!        ***  ensure that  h  is neither too small nor too big  ***

h=MAX(h,hmin*axibar)
IF (h >= hmax0*axibar) h=axibar*h0**(two/three)

!        ***  compute central difference  ***

irc=-i
GO TO 100

50 h=-w(hsave)
i=ABS(irc)
IF (h > zero) GO TO 60
w(fh)=fx
GO TO 100

60 g(i)=(w(fh)-fx)/(two*h)
x(i)=w(xisave)
GO TO 20

!     ***  compute forward differences in various cases  ***

70 IF (h >= hmax0*axibar) h=h0*axibar
IF (alphai*gi < zero) h=-h
GO TO 100
80 h=axibar
GO TO 100
90 h=h0*axibar

100 x(i)=w(xisave)+h
w(hsave)=h
GO TO 130

!     ***  compute actual forward difference  ***

110 g(irc)=(fx-w(fx0))/w(hsave)
x(irc)=w(xisave)
GO TO 20

!  ***  restore fx and indicate that g has been computed  ***

120 fx=w(fx0)
irc=0

130 RETURN
!  ***  last card of sgrad2 follows  ***
END SUBROUTINE sgrad2


SUBROUTINE vvmulp (n,x,y,z,k)

! ***  set x(i) = y(i) * z(i)**k, 1 <= i <= n (for k = 1 or -1)  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: n,k
REAL (dp), INTENT(IN)  :: y(:),z(:)
REAL (dp), INTENT(OUT) :: x(:)

IF (k < 0) THEN
  x(1:n)=y(1:n)/z(1:n)
ELSE
  x(1:n)=y(1:n)*z(1:n)
END IF

RETURN
!  ***  last card of vvmulp follows  ***
END SUBROUTINE vvmulp


SUBROUTINE wzbfgs (l,n,s,w,y,z)

!  ***  compute  y  and  z  for  lupdat  corresponding to bfgs update.

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: n
REAL (dp), INTENT(IN)     :: s(:),y(:)
REAL (dp), INTENT(IN OUT) :: l(:)
REAL (dp), INTENT(OUT)    :: w(:),z(:)
!     dimension l(n*(n+1)/2)

!--------------------------  parameter usage  --------------------------

! l (i/o) cholesky factor of hessian, a lower triang. matrix stored
!             compactly by rows.
! n (input) order of  l  and length of  s,  w,  y,  z.
! s (input) the step just taken.
! w (output) right singular vector of rank 1 correction to l.
! y (input) change in gradients corresponding to s.
! z (output) left singular vector of rank 1 correction to l.

!-------------------------------  notes  -------------------------------

!  ***  algorithm notes  ***

!        when  s  is computed in certain ways, e.g. by  gqtstp  or
!     dbldog,  it is possible to save n**2/2 operations since  (l**t)*s
!     or  l*(l**t)*s is then known.
!        if the bfgs update to l*(l**t) would reduce its determinant to
!     less than eps times its old value, then this routine in effect
!     replaces  y  by  theta*y + (1 - theta)*l*(l**t)*s,  where  theta
!     (between 0 and 1) is chosen to make the reduction factor = eps.

!  ***  general  ***

!     coded by david m. gay (fall 1979).
!     this subroutine was written in connection with research supported
!     by the national science foundation under grants mcs-7600324 and
!     mcs-7906671.

!------------------------  external quantities  ------------------------

!  ***  functions and subroutines called  ***

! dotprd returns inner product of two vectors.
! livmul multiplies l**-1 times a vector.
! ltvmul multiplies l**t times a vector.

!--------------------------  local variables  --------------------------

INTEGER   :: i
REAL (dp) :: cs, cy, eps=0.1_dp, epsrt, one=1._dp, shs, ys, theta

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

CALL ltvmul (n,w,l,s)
shs=dotprd(n,w,w)
ys=dotprd(n,y,s)
IF (ys >= eps*shs) GO TO 10
theta=(one-eps)*shs/(shs-ys)
epsrt=SQRT(eps)
cy=theta/(shs*epsrt)
cs=(one + (theta-one)/epsrt)/shs
GO TO 20
10 cy=one/(SQRT(ys)*SQRT(shs))
cs=one/shs
20 CALL livmul (n,z,l,y)
DO i=1,n
  z(i)=cy*z(i)-cs*w(i)
END DO

RETURN
!  ***  last card of wzbfgs follows  ***
END SUBROUTINE wzbfgs


SUBROUTINE assst (iv,v)

!  ***  assess candidate step (***sol version 2.3)  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN OUT)   :: iv(:)
REAL (dp), INTENT(IN OUT) :: v(:)

!  ***  purpose  ***

!      this subroutine is called by an unconstrained minimization routine
!   to assess the next candidate step.  it may recommend one of several
!   courses of action, such as accepting the step, recomputing it using
!   the same or a new quadratic model, or halting due to convergence or
!   false convergence.   See the return code listing below.

!--------------------------  parameter usage  --------------------------

!  iv (i/o) integer parameter and scratch vector -- see description
!             below of iv values referenced.
! liv (in)  length of iv array.
!  lv (in)  length of v array.
!   v (i/o) real parameter and scratch vector -- see description
!             below of v values referenced.

!  ***  iv values referenced  ***

!    iv(irc) (i/o) on input for the first step tried in a new iteration,
!             iv(irc) should be set to 3 or 4 (the value to which it is
!             set when step is definitely to be accepted).  on input
!             after step has been recomputed, iv(irc) should be
!             unchanged since the previous return of assst.
!                on output, iv(irc) is a return code having one of the
!             following values...
!                  1 = switch models or try smaller step.
!                  2 = switch models or accept step.
!                  3 = accept step and determine v(radfac) by gradient
!                       tests.
!                  4 = accept step, v(radfac) has been determined.
!                  5 = recompute step (using the same model).
!                  6 = recompute step with radius = v(lmaxs) but do not
!                       evaulate the objective function.
!                  7 = x-convergence (see v(xctol)).
!                  8 = relative function convergence (see v(rfctol)).
!                  9 = both x- and relative function convergence.
!                 10 = absolute function convergence (see v(afctol)).
!                 11 = singular convergence (see v(lmaxs)).
!                 12 = false convergence (see v(xftol)).
!                 13 = iv(irc) was out of range on input.
!             return code i has precdence over i+1 for i = 9, 10, 11.
! iv(mlstgd) (i/o) saved value of iv(model).
!  iv(model) (i/o) on input, iv(model) should be an integer identifying
!             the current quadratic model of the objective function.
!             if a previous step yielded a better function reduction,
!             then iv(model) will be set to iv(mlstgd) on output.
! iv(nfcall) (in)  invocation count for the objective function.
! iv(nfgcal) (i/o) value of iv(nfcall) at step that gave the biggest
!             function reduction this iteration.  iv(nfgcal) remains
!             unchanged until a function reduction is obtained.
! iv(radinc) (i/o) the number of radius increases (or minus the number
!             of decreases) so far this iteration.
! iv(restor) (out) set to 1 if v(f) has been restored and x should be
!             restored to its initial value, to 2 if x should be saved,
!             to 3 if x should be restored from the saved value, and to
!             0 otherwise.
!  iv(stage) (i/o) count of the number of models tried so far in the
!             current iteration.
! iv(stglim) (in)  maximum number of models to consider.
! iv(switch) (out) set to 0 unless a new model is being tried and it
!             gives a smaller function value than the previous model,
!             in which case assst sets iv(switch) = 1.
! iv(toobig) (in)  is nonzero if step was too big (e.g. if it caused
!             overflow).
!   iv(xirc) (i/o) value that iv(irc) would have in the absence of
!             convergence, false convergence, and oversized steps.

!  ***  v values referenced  ***

! v(afctol) (in)  absolute function convergence tolerance.  if the
!             absolute value of the current function value v(f) is less
!             than v(afctol), then assst returns with iv(irc) = 10.
! v(decfac) (in)  factor by which to decrease radius when iv(toobig) is
!             nonzero.
! v(dstnrm) (in)  the 2-norm of d*step.
! v(dstsav) (i/o) value of v(dstnrm) on saved step.
!   v(dst0) (in)  the 2-norm of d times the newton step (when defined,
!             i.e., for v(nreduc) .ge. 0).
!      v(f) (i/o) on both input and output, v(f) is the objective func-
!             tion value at x.  if x is restored to a previous value,
!             then v(f) is restored to the corresponding value.
!   v(fdif) (out) the function reduction v(f0) - v(f) (for the output
!             value of v(f) if an earlier step gave a bigger function
!             decrease, and for the input value of v(f) otherwise).
! v(flstgd) (i/o) saved value of v(f).
!     v(f0) (in)  objective function value at start of iteration.
! v(gtslst) (i/o) value of v(gtstep) on saved step.
! v(gtstep) (in)  inner product between step and gradient.
! v(incfac) (in)  minimum factor by which to increase radius.
!  v(lmaxs) (in)  maximum reasonable step size (and initial step bound).
!             if the actual function decrease is no more than twice
!             what was predicted, if a return with iv(irc) = 7, 8, 9,
!             or 10 does not occur, if v(dstnrm) .gt. v(lmaxs), and if
!             v(preduc) .le. v(sctol) * abs(v(f0)), then assst re-
!             turns with iv(irc) = 11.  if so doing appears worthwhile,
!             then assst repeats this test with v(preduc) computed for
!             a step of length v(lmaxs) (by a return with iv(irc) = 6).
! v(nreduc) (i/o)  function reduction predicted by quadratic model for
!             newton step.  if assst is called with iv(irc) = 6, i.e.,
!             if v(preduc) has been computed with radius = v(lmaxs) for
!             use in the singular convervence test, then v(nreduc) is
!             set to -v(preduc) before the latter is restored.
! v(plstgd) (i/o) value of v(preduc) on saved step.
! v(preduc) (i/o) function reduction predicted by quadratic model for
!             current step.
! v(radfac) (out) factor to be used in determining the new radius,
!             which should be v(radfac)*dst, where  dst  is either the
!             output value of v(dstnrm) or the 2-norm of
!             diag(newd)*step  for the output value of step and the
!             updated version, newd, of the scale vector d.  for
!             iv(irc) = 3, v(radfac) = 1.0 is returned.
! v(rdfcmn) (in)  minimum value for v(radfac) in terms of the input
!             value of v(dstnrm) -- suggested value = 0.1.
! v(rdfcmx) (in)  maximum value for v(radfac) -- suggested value = 4.0.
!  v(reldx) (in) scaled relative change in x caused by step, computed
!             (e.g.) by function  reldst  as
!                 max (d(i)*abs(x(i)-x0(i)), 1 .le. i .le. p) /
!                    max (d(i)*(abs(x(i))+abs(x0(i))), 1 .le. i .le. p).
! v(rfctol) (in)  relative function convergence tolerance.  if the
!             actual function reduction is at most twice what was pre-
!             dicted and  v(nreduc) .le. v(rfctol)*abs(v(f0)),  then
!             assst returns with iv(irc) = 8 or 9.
! v(stppar) (in)  marquardt parameter -- 0 means full newton step.
! v(tuner1) (in)  tuning constant used to decide if the function
!             reduction was much less than expected.  suggested
!             value = 0.1.
! v(tuner2) (in)  tuning constant used to decide if the function
!             reduction was large enough to accept step.  suggested
!             value = 10**-4.
! v(tuner3) (in)  tuning constant used to decide if the radius
!             should be increased.  suggested value = 0.75.
!  v(xctol) (in)  x-convergence criterion.  if step is a newton step
!             (v(stppar) = 0) having v(reldx) .le. v(xctol) and giving
!             at most twice the predicted function decrease, then
!             assst returns iv(irc) = 7 or 9.
!  v(xftol) (in)  false convergence tolerance.  if step gave no or only
!             a small function decrease and v(reldx) .le. v(xftol),
!             then assst returns with iv(irc) = 12.

!-------------------------------  notes  -------------------------------

!  ***  application and usage restrictions  ***

!        this routine is called as part of the nl2sol (nonlinear
!     least-squares) package.  it may be used in any unconstrained
!     minimization solver that uses dogleg, goldfeld-quandt-trotter,
!     or levenberg-marquardt steps.

!  ***  algorithm notes  ***

!        see (1) for further discussion of the assessing and model
!     switching strategies.  while nl2sol considers only two models,
!     assst is designed to handle any number of models.

!  ***  usage notes  ***

!        on the first call of an iteration, only the i/o variables
!     step, x, iv(irc), iv(model), v(f), v(dstnrm), v(gtstep), and
!     v(preduc) need have been initialized.  between calls, no i/o
!     values execpt step, x, iv(model), v(f) and the stopping toler-
!     ances should be changed.
!        after a return for convergence or false convergence, one can
!     change the stopping tolerances and call assst again, in which
!     case the stopping tests will be repeated.

!  ***  references  ***

!     (1) dennis, j.e., jr., gay, d.m., and welsch, r.e. (1981),
!        an adaptive nonlinear least-squares algorithm,
!        acm trans. math. software, vol. 7, no. 3.

!     (2) powell, m.j.d. (1970)  a fortran subroutine for solving
!        systems of nonlinear algebraic equations, in numerical
!        methods for nonlinear algebraic equations, edited by
!        p. rabinowitz, gordon and breach, london.

!  ***  history  ***

!        john dennis designed much of this routine, starting with
!     ideas in (2). roy welsch suggested the model switching strategy.
!        david gay and stephen peters cast this subroutine into a more
!     portable form (winter 1977), and david gay cast it into its
!     present form (fall 1978).

!  ***  general  ***

!     this subroutine was written in connection with research
!     supported by the national science foundation under grants
!     mcs-7600324, dcr75-10143, 76-14311dss, mcs76-11989, and
!     mcs-7906671.

!------------------------  external quantities  ------------------------

!  ***  no external functions and subroutines  ***

!  ***  no common blocks  ***

!--------------------------  local variables  --------------------------

LOGICAL   :: goodx
INTEGER   :: i,nfc
REAL (dp) :: emax,emaxs,gts,rfac1,xmax

REAL (dp) :: half=0.5D+0, one=1.d+0, onep2=1.2D+0, two=2.d+0, zero=0.d+0

!  ***  subscripts for iv and v  ***

INTEGER :: irc=29, mlstgd=32, model=5, nfcall=6, nfgcal=7, radinc=8,   &
           restor=9, stage=10, stglim=11, switch=12, toobig=2, xirc=13
INTEGER :: afctol=31, decfac=22, dstnrm=2, dst0=3, dstsav=18, f=10, fdif=11,  &
           flstgd=12, f0=13, gtslst=14, gtstep=4, incfac=23, lmaxs=36,   &
           nreduc=6, plstgd=15, preduc=7, radfac=16, rdfcmn=24, rdfcmx=25,   &
           reldx=17, rfctol=32, sctol=37, stppar=5, tuner1=26, tuner2=27,   &
           tuner3=28, xctol=33, xftol=34

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

nfc=iv(nfcall)
iv(switch)=0
iv(restor)=0
rfac1=one
goodx=.true.
i=iv(irc)
SELECT CASE (i)
  CASE (1)
    GO TO 20
  CASE (2)
    GO TO 30
  CASE (3:4)
    GO TO 10
  CASE (5)
    GO TO 40
  CASE (6)
    GO TO 260
  CASE (7:11)
    GO TO 200
  CASE (12)
    GO TO 150
  CASE DEFAULT
    iv(irc)=13
    GO TO 280
END SELECT

!  ***  initialize for new iteration  ***

10 iv(stage)=1
iv(radinc)=0
v(flstgd)=v(f0)
IF (iv(toobig) == 0) GO TO 90
iv(stage)=-1
iv(xirc)=i
GO TO 60

!  ***  step was recomputed with new model or smaller radius  ***
!  ***  first decide which  ***

20 IF (iv(model) /= iv(mlstgd)) GO TO 30
!        ***  old model retained, smaller radius tried  ***
!        ***  do not consider any more new models this iteration  ***
iv(stage)=iv(stglim)
iv(radinc)=-1
GO TO 90

!  ***  a new model is being tried.  decide whether to keep it.  ***

30 iv(stage)=iv(stage) + 1

!     ***  now we add the possibiltiy that step was recomputed with  ***
!     ***  the same model, perhaps because of an oversized step.     ***

40 IF (iv(stage) > 0) GO TO 50

!        ***  step was recomputed because it was too big.  ***

IF (iv(toobig) /= 0) GO TO 60

!        ***  restore iv(stage) and pick up where we left off.  ***

iv(stage)=-iv(stage)
i=iv(xirc)

SELECT CASE (i)
  CASE (1)
    GO TO 20
  CASE (2)
    GO TO 30
  CASE (3:4)
    GO TO 90
  CASE DEFAULT
    GO TO 70
END SELECT

50 IF (iv(toobig) == 0) GO TO 70

!  ***  handle oversize step  ***

IF (iv(radinc) > 0) GO TO 80
iv(stage)=-iv(stage)
iv(xirc)=iv(irc)

60 v(radfac)=v(decfac)
iv(radinc)=iv(radinc)-1
iv(irc)=5
iv(restor)=1
GO TO 280

70 IF (v(f) < v(flstgd)) GO TO 90

!     *** the new step is a loser.  restore old model.  ***

IF (iv(model) == iv(mlstgd)) GO TO 80
iv(model)=iv(mlstgd)
iv(switch)=1

!     ***  restore step, etc. only if a previous step decreased v(f).

80 IF (v(flstgd) >= v(f0)) GO TO 90
iv(restor)=1
v(f)=v(flstgd)
v(preduc)=v(plstgd)
v(gtstep)=v(gtslst)
IF (iv(switch) == 0) rfac1=v(dstnrm)/v(dstsav)
v(dstnrm)=v(dstsav)
nfc=iv(nfgcal)
goodx=.false.

90 v(fdif)=v(f0)-v(f)
IF (v(fdif) > v(tuner2)*v(preduc)) GO TO 120
IF (iv(radinc) > 0) GO TO 120

!        ***  no (or only a trivial) function decrease
!        ***  -- so try new model or smaller radius

IF (v(f) < v(f0)) GO TO 100
iv(mlstgd)=iv(model)
v(flstgd)=v(f)
v(f)=v(f0)
iv(restor)=1
GO TO 110
100 iv(nfgcal)=nfc
110 iv(irc)=1
IF (iv(stage) < iv(stglim)) GO TO 140
iv(irc)=5
iv(radinc)=iv(radinc)-1
GO TO 140

!  ***  nontrivial function decrease achieved  ***

120 iv(nfgcal)=nfc
rfac1=one
v(dstsav)=v(dstnrm)
IF (v(fdif) > v(preduc)*v(tuner1)) GO TO 170

!  ***  decrease was much less than predicted -- either change models
!  ***  or accept step with decreased radius.

IF (iv(stage) >= iv(stglim)) GO TO 130
!        ***  consider switching models  ***
iv(irc)=2
GO TO 140

!     ***  accept step with decreased radius  ***

130 iv(irc)=4

!  ***  set v(radfac) to fletcher*s decrease factor  ***

140 iv(xirc)=iv(irc)
emax=v(gtstep) + v(fdif)
v(radfac)=half*rfac1
IF (emax < v(gtstep)) v(radfac)=rfac1*MAX(v(rdfcmn),half*  &
v(gtstep)/emax)

!  ***  do false convergence test  ***

150 IF (v(reldx) <= v(xftol)) GO TO 160
iv(irc)=iv(xirc)
IF (v(f) < v(f0)) GO TO 180
GO TO 210

160 iv(irc)=12
GO TO 220

!  ***  handle good function decrease  ***

170 IF (v(fdif) < (-v(tuner3)*v(gtstep))) GO TO 190

!     ***  increasing radius looks worthwhile.  see if we just
!     ***  recomputed step with a decreased radius or restored step
!     ***  after recomputing it with a larger radius.

IF (iv(radinc) < 0) GO TO 190
IF (iv(restor) == 1) GO TO 190

!        ***  we did not.  try a longer step unless this was a newton
!        ***  step.

v(radfac)=v(rdfcmx)
gts=v(gtstep)
IF (v(fdif) < (half/v(radfac)-one)*gts) v(radfac)=MAX(v(incfac)  &
,half*gts/(gts + v(fdif)))
iv(irc)=4
IF (v(stppar) == zero) GO TO 210
IF (v(dst0) >= zero .AND. (v(dst0) < two*v(dstnrm) .OR. v(nreduc) <  &
onep2*v(fdif))) GO TO 210
!             ***  step was not a newton step.  recompute it with
!             ***  a larger radius.
iv(irc)=5
iv(radinc)=iv(radinc) + 1

!  ***  save values corresponding to good step  ***

180 v(flstgd)=v(f)
iv(mlstgd)=iv(model)
IF (iv(restor) /= 1) iv(restor)=2
v(dstsav)=v(dstnrm)
iv(nfgcal)=nfc
v(plstgd)=v(preduc)
v(gtslst)=v(gtstep)
GO TO 210

!  ***  accept step with radius unchanged  ***

190 v(radfac)=one
iv(irc)=3
GO TO 210

!  ***  come here for a restart after convergence  ***

200 iv(irc)=iv(xirc)
IF (v(dstsav) >= zero) GO TO 220
iv(irc)=12
GO TO 220

!  ***  perform convergence tests  ***

210 iv(xirc)=iv(irc)
220 IF (iv(restor) == 1 .AND. v(flstgd) < v(f0)) iv(restor)=3
IF (ABS(v(f)) < v(afctol)) iv(irc)=10
IF (half*v(fdif) > v(preduc)) GO TO 280
emax=v(rfctol)*ABS(v(f0))
emaxs=v(sctol)*ABS(v(f0))
IF (v(dstnrm) > v(lmaxs) .AND. v(preduc) <= emaxs) iv(irc)=11
IF (v(dst0) < zero) GO TO 230
i=0
IF ((v(nreduc) > zero .AND. v(nreduc) <= emax) .OR. (v(nreduc) ==  &
zero .AND. v(preduc) == zero)) i=2
IF (v(stppar) == zero .AND. v(reldx) <= v(xctol) .AND. goodx) i=i + 1
IF (i > 0) iv(irc)=i + 6

!  ***  consider recomputing step of length v(lmaxs) for singular
!  ***  convergence test.

230 IF (iv(irc) > 5 .AND. iv(irc) /= 12) GO TO 280
IF (v(dstnrm) > v(lmaxs)) GO TO 240
IF (v(preduc) >= emaxs) GO TO 280
IF (v(dst0) <= zero) GO TO 250
IF (half*v(dst0) <= v(lmaxs)) GO TO 280
GO TO 250
240 IF (half*v(dstnrm) <= v(lmaxs)) GO TO 280
xmax=v(lmaxs)/v(dstnrm)
IF (xmax*(two-xmax)*v(preduc) >= emaxs) GO TO 280
250 IF (v(nreduc) < zero) GO TO 270

!  ***  recompute v(preduc) for use in singular convergence test  ***

v(gtslst)=v(gtstep)
v(dstsav)=v(dstnrm)
IF (iv(irc) == 12) v(dstsav)=-v(dstsav)
v(plstgd)=v(preduc)
i=iv(restor)
iv(restor)=2
IF (i == 3) iv(restor)=0
iv(irc)=6
GO TO 280

!  ***  perform singular convergence test with recomputed v(preduc)  ***

260 v(gtstep)=v(gtslst)
v(dstnrm)=ABS(v(dstsav))
iv(irc)=iv(xirc)
IF (v(dstsav) <= zero) iv(irc)=12
v(nreduc)=-v(preduc)
v(preduc)=v(plstgd)
iv(restor)=3
270 IF (-v(nreduc) <= v(rfctol)*ABS(v(f0))) iv(irc)=11

280 RETURN

!  ***  last card of assst follows  ***
END SUBROUTINE assst


SUBROUTINE deflt (alg,iv,liv,lv,v)

!  ***  supply ***sol (version 2.3) default values to iv and v  ***

!  ***  alg = 1 means regression constants.
!  ***  alg = 2 means general unconstrained optimization constants.

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: liv,lv
INTEGER, INTENT(IN)    :: alg
INTEGER, INTENT(OUT)   :: iv(:)
REAL (dp), INTENT(OUT) :: v(:)

! vdflt.... provides default values to v.

INTEGER :: miv,mv

!  ***  iv subscript values  ***

INTEGER :: algsav=51, covprt=14, covreq=15, dtype=16, hc=71, ierr=75,   &
           inith=25, inits=25, ipivot=76, ivneed=3, lastiv=44, lastv=45,   &
           lmat=42, mxfcal=17, mxiter=18, nfcov=52, ngcov=53, nvdflt=50,   &
           outlev=19, parprt=20, parsav=49, perm=58, prunit=21, qrtyp=80,   &
           rdreq=57, rmat=78, solprt=22, statpr=23, vneed=4, vsave=60,   &
           x0prt=24
INTEGER :: miniv(2) = (/80, 59/), minv(2) = (/98, 71/)

!-------------------------------  body  --------------------------------

IF (alg < 1 .OR. alg > 2) GO TO 40
miv=miniv(alg)
IF (liv < miv) GO TO 20
mv=minv(alg)
IF (lv < mv) GO TO 30
CALL vdflt (alg,v)
iv(1)=12
iv(algsav)=alg
iv(ivneed)=0
iv(lastiv)=miv
iv(lastv)=mv
iv(lmat)=mv + 1
iv(mxfcal)=200
iv(mxiter)=150
iv(outlev)=1
iv(parprt)=1
iv(perm)=miv + 1
iv(prunit)=imdcon(1)
iv(solprt)=1
iv(statpr)=1
iv(vneed)=0
iv(x0prt)=1

IF (alg >= 2) GO TO 10

!  ***  regression  values

iv(covprt)=3
iv(covreq)=1
iv(dtype)=1
iv(hc)=0
iv(ierr)=0
iv(inits)=0
iv(ipivot)=0
iv(nvdflt)=32
iv(parsav)=67
iv(qrtyp)=1
iv(rdreq)=3
iv(rmat)=0
iv(vsave)=58
GO TO 50

!  ***  general optimization values

10 iv(dtype)=0
iv(inith)=1
iv(nfcov)=0
iv(ngcov)=0
iv(nvdflt)=25
iv(parsav)=47
GO TO 50

20 iv(1)=15
GO TO 50

30 iv(1)=16
GO TO 50

40 iv(1)=67

50 RETURN
!  ***  last card of deflt follows  ***
END SUBROUTINE deflt


FUNCTION dotprd (p,x,y) RESULT(fn_val)

!  ***  return the inner product of the p-vectors x and y.  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)   :: p
REAL (dp), INTENT(IN) :: x(:),y(:)
REAL (dp)             :: fn_val

INTEGER   :: i
REAL (dp) :: one=1._dp, sqteta = 0._dp, t, zero = 0._dp

fn_val=zero
IF (p <= 0) RETURN
IF (sqteta == zero) sqteta=rmdcon(2)
DO i=1,p
  t=MAX(ABS(x(i)),ABS(y(i)))
  IF (t < one) THEN
    IF (t < sqteta) CYCLE
    t=(x(i)/sqteta)*y(i)
    IF (ABS(t) < sqteta) CYCLE
  END IF
  fn_val=fn_val + x(i)*y(i)
END DO

RETURN
!  ***  last card of dotprd follows  ***
END FUNCTION dotprd


SUBROUTINE itsum (d,g,iv,p,v,x)

!  ***  print iteration summary for ***sol (version 2.3)  ***

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: p
INTEGER, INTENT(IN OUT)   :: iv(:)
REAL (dp), INTENT(IN)     :: d(:),g(:),x(:)
REAL (dp), INTENT(IN OUT) :: v(:)

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

!  ***  local variables  ***

INTEGER   :: alg,i,iv1,m,nf,ng,ol,pu
REAL (dp) :: nreldf,oldf,preldf,reldf,zero=0.d+0

!  ***  no external functions or subroutines  ***

!  ***  iv subscript values  ***

INTEGER :: algsav=51, needhd=36, nfcall=6, nfcov=52, ngcall=30, ngcov=53,   &
           niter=31, outlev=19, prntit=39, prunit=21, solprt=22, statpr=23,  &
           sused=64, x0prt=24

!  ***  v subscript values  ***

INTEGER :: dstnrm=2, f=10, f0=13, fdif=11, nreduc=6, preduc=7, reldx=17,  &
           stppar=5

CHARACTER (LEN=4) :: model1(6) = (/ '    ', '    ', '    ',  &
                                    '    ', '  g ', '  s ' /),  &
                     model2(6) = (/ ' g  ', ' s  ', 'g-s ',  &
                                    's-g ', '-s-g', '-g-s' /)

!-------------------------------  body  --------------------------------

pu=iv(prunit)
IF (pu == 0) GO TO 540
iv1=iv(1)
IF (iv1 > 62) iv1=iv1-51
ol=iv(outlev)
alg=iv(algsav)
IF (iv1 < 2 .OR. iv1 > 15) GO TO 370
IF (iv1 >= 12) GO TO 120
IF (iv1 == 2 .AND. iv(niter) == 0) GO TO 390
IF (ol == 0) GO TO 120
IF (iv1 >= 10 .AND. iv(prntit) == 0) GO TO 120
IF (iv1 > 2) GO TO 10
iv(prntit)=iv(prntit) + 1
IF (iv(prntit) < ABS(ol)) GO TO 540
10 nf=iv(nfcall)-ABS(iv(nfcov))
iv(prntit)=0
reldf=zero
preldf=zero
oldf=MAX(ABS(v(f0)),ABS(v(f)))
IF (oldf <= zero) GO TO 20
reldf=v(fdif)/oldf
preldf=v(preduc)/oldf
20 IF (ol > 0) GO TO 60

!        ***  print short summary line  ***

IF (iv(needhd) == 1 .AND. alg == 1) WRITE (pu,30)
30 FORMAT (/'   it   nf','      f','       reldf','   preldf','   reldx',  &
            '  model  stppar')
IF (iv(needhd) == 1 .AND. alg == 2) WRITE (pu,40)
40 FORMAT (/'    it   nf','       f','        reldf','    preldf','    reldx',  &
            '   stppar')
iv(needhd)=0
IF (alg == 2) GO TO 50
m=iv(sused)
WRITE (pu,100) iv(niter),nf,v(f),reldf,preldf,v(reldx),model1(m),  &
               model2(m),v(stppar)
GO TO 120

50 WRITE (pu,110) iv(niter),nf,v(f),reldf,preldf,v(reldx),v(stppar)
GO TO 120

!     ***  print long summary line  ***

60 IF (iv(needhd) == 1 .AND. alg == 1) WRITE (pu,70)
70 FORMAT (/'    it   nf','      f','       reldf','   preldf','   reldx',  &
            '  model  stppar','  d*step','  npreldf')
IF (iv(needhd) == 1 .AND. alg == 2) WRITE (pu,80)
80 FORMAT (/'    it   nf','       f','        reldf','    preldf','    reldx',  &
            '   stppar','   d*step','   npreldf')
iv(needhd)=0
nreldf=zero
IF (oldf > zero) nreldf=v(nreduc)/oldf
IF (alg == 2) GO TO 90
m=iv(sused)
WRITE (pu,100) iv(niter),nf,v(f),reldf,preldf,v(reldx),model1(m),  &
               model2(m),v(stppar),v(dstnrm),nreldf
GO TO 120

90 WRITE (pu,110) iv(niter),nf,v(f),reldf,preldf,v(reldx),v(stppar),  &
                  v(dstnrm),nreldf
100 FORMAT (i6,i5,f10.3,2f9.2,f8.1,a3,a4,2f8.1,f9.2)
110 FORMAT (i6,i5,f11.3,2f10.2,3f9.1,f10.2)

120 IF (iv(statpr) < 0) GO TO 430

SELECT CASE (iv1)
  CASE (1:2)
    GO TO 540
  CASE (3)
    WRITE (pu,140)
    140 FORMAT (/' ***** x-convergence *****')
    GO TO 430
  CASE (4)
    WRITE (pu,160)
    160 FORMAT (/' ***** relative function convergence *****')
    GO TO 430
  CASE (5)
    WRITE (pu,180)
    180 FORMAT (/' ***** x- and relative function convergence *****')
    GO TO 430
  CASE (6)
    WRITE (pu,200)
    200 FORMAT (/' ***** absolute function convergence *****')
    GO TO 430
  CASE (7)
    WRITE (pu,220)
    220 FORMAT (/' ***** singular convergence *****')
    GO TO 430
  CASE (8)
    WRITE (pu,240)
    240 FORMAT (/' ***** false convergence *****')
    GO TO 430
  CASE (9)
    WRITE (pu,260)
    260 FORMAT (/' ***** function evaluation limit *****')
    GO TO 430
  CASE (10)
    WRITE (pu,280)
    280 FORMAT (/' ***** iteration limit *****')
    GO TO 430
  CASE (11)
    WRITE (pu,300)
    300 FORMAT (/' ***** stopx *****')
    GO TO 430
  CASE (12)
    WRITE (pu,320)
    320 FORMAT (/' ***** initial f(x) cannot be computed *****')
    GO TO 390
  CASE (13)
    WRITE (pu,340)
    340 FORMAT (/' ***** bad parameters to assess *****')
    GO TO 540
  CASE (14)
    WRITE (pu,360)
    360 FORMAT (/' ***** gradient could not be computed *****')
    IF (iv(niter) > 0) GO TO 480
    GO TO 390
  CASE (15)
    GO TO 520
  CASE DEFAULT
    GO TO 430
END SELECT

370 WRITE (pu,380) iv(1)
380 FORMAT (/' ***** iv(1) =',i5,' *****')
GO TO 540

!  ***  initial call on itsum  ***

390 IF (iv(x0prt) /= 0) WRITE (pu,400) (i,x(i),d(i),i=1,p)
400 FORMAT (/'     i     initial x(i)','        d(i)'//(' ',i5,f17.6,f14.3))
!     *** the following are to avoid undefined variables when the
!     *** function evaluation limit is 1...
v(dstnrm)=zero
v(fdif)=zero
v(nreduc)=zero
v(preduc)=zero
v(reldx)=zero
IF (iv1 >= 12) GO TO 540
iv(needhd)=0
iv(prntit)=0
IF (ol == 0) GO TO 540
IF (ol < 0 .AND. alg == 1) WRITE (pu,30)
IF (ol < 0 .AND. alg == 2) WRITE (pu,40)
IF (ol > 0 .AND. alg == 1) WRITE (pu,70)
IF (ol > 0 .AND. alg == 2) WRITE (pu,80)
IF (alg == 1) WRITE (pu,410) v(f)
IF (alg == 2) WRITE (pu,420) v(f)
410 FORMAT (/'     0 1',f10.3)
420 FORMAT (/'     0 1',f11.3)
GO TO 540

!  ***  print various information requested on solution  ***

430 iv(needhd)=1
IF (iv(statpr) == 0) GO TO 480
oldf=MAX(ABS(v(f0)),ABS(v(f)))
preldf=zero
nreldf=zero
IF (oldf <= zero) GO TO 440
preldf=v(preduc)/oldf
nreldf=v(nreduc)/oldf
440 nf=iv(nfcall)-iv(nfcov)
ng=iv(ngcall)-iv(ngcov)
WRITE (pu,450) v(f),v(reldx),nf,ng,preldf,nreldf
450 FORMAT (/' FUNCTION', f17.6, '   reldx', f17.3/ ' func. evals', i8,  &
             '          grad. evals', i8/ ' preldf', f16.3, '      npreldf', &
             f15.3)
IF (iv(nfcov) > 0) WRITE (pu,460) iv(nfcov)
460 FORMAT (/' ',i4,' extra func. evals for covariance and diagnostics.')
IF (iv(ngcov) > 0) WRITE (pu,470) iv(ngcov)
470 FORMAT (' ',i4,' extra grad. evals for covariance and diagnostics.')

480 IF (iv(solprt) == 0) GO TO 540
iv(needhd)=1
WRITE (pu,490)
490 FORMAT (/'     i      final x(i)','        d(i)','          g(i)'/)
DO i=1,p
  WRITE (pu,510) i,x(i),d(i),g(i)
END DO
510 FORMAT (' ',i5,f16.6,2f14.3)
GO TO 540

520 WRITE (pu,530)
530 FORMAT (/' inconsistent dimensions')
540 RETURN
!  ***  last card of itsum follows  ***
END SUBROUTINE itsum


SUBROUTINE litvmu (n,x,l,y)

!  ***  solve  (l**t)*x = y,  where  l  is an  n x n  lower triangular
!  ***  matrix stored compactly by rows.  x and y may occupy the same
!  ***  storage.  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: n
REAL (dp), INTENT(IN)  :: l(:),y(:)
REAL (dp), INTENT(OUT) :: x(:)

! Local variables
INTEGER   :: i,ii,ij,im1,i0,j,np1
REAL (dp) :: xi,zero=0._dp

DO i=1,n
  x(i)=y(i)
END DO
np1=n + 1
i0=n*(n+1)/2
DO ii=1,n
  i=np1-ii
  xi=x(i)/l(i0)
  x(i)=xi
  IF (i <= 1) GO TO 40
  i0=i0-i
  IF (xi == zero) CYCLE
  im1=i-1
  DO j=1,im1
    ij=i0 + j
    x(j)=x(j) - xi*l(ij)
  END DO
END DO
40 RETURN
!  ***  last card of litvmu follows  ***
END SUBROUTINE litvmu


SUBROUTINE livmul (n,x,l,y)

!  ***  solve  l*x = y, where  l  is an  n x n  lower triangular
!  ***  matrix stored compactly by rows.  x and y may occupy the same
!  ***  storage.  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: n
REAL (dp), INTENT(IN)  :: l(:),y(:)
REAL (dp), INTENT(OUT) :: x(:)

! Local variables
INTEGER   :: i,j,k
REAL (dp) :: t,zero=0._dp

DO k=1,n
  IF (y(k) /= zero) GO TO 20
  x(k)=zero
END DO
GO TO 40
20 j=k*(k+1)/2
x(k)=y(k)/l(j)
IF (k >= n) GO TO 40
k=k + 1
DO i=k,n
  t=dotprd(i-1,l(j+1:),x)
  j=j + i
  x(i)=(y(i)-t)/l(j)
END DO
40 RETURN
!  ***  last card of livmul follows  ***
END SUBROUTINE livmul


SUBROUTINE parck (alg,d,iv,liv,lv,n,v)

!  ***  check ***sol (version 2.3) parameters, print changed values  ***

!  ***  alg = 1 for regression, alg = 2 for general unconstrained opt.

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: alg,liv,lv,n
INTEGER, INTENT(IN OUT)   :: iv(:)
REAL (dp), INTENT(IN OUT) :: d(:),v(:)

! vcopy  -- copies one vector to another.
! vdflt  -- supplies default parameter values to v alone.

!  ***  local variables  ***

INTEGER           :: i,ii,iv1,j,k,l,m,miv1,miv2,ndfalt,parsv1,pu
CHARACTER (LEN=4) :: which(3)
REAL (dp)         :: vk

!  ***  iv and v subscripts  ***

INTEGER :: algsav=51, dinit=38, dtype=16, dtype0=54, epslon=19, inits=25,  &
           ivneed=3, lastiv=44, lastv=45, lmat=42, nextiv=46, nextv=47,   &
           nvdflt=50, oldn=38, parprt=20, parsav=49, perm=58, prunit=21,  &
           vneed=4

REAL (dp) :: big=0.d+0, machep=-1.d+0, tiny=1.d+0, zero=0.d+0

CHARACTER (LEN=6) :: vn(34) = (/ 'epslon', 'phmnfc', 'phmxfc', 'decfac',  &
                                 'incfac', 'rdfcmn', 'rdfcmx', 'tuner1',  &
                                 'tuner2', 'tuner3', 'tuner4', 'tuner5',  &
                                 'afctol', 'rfctol', 'xctol.', 'xftol.',  &
                                 'lmax0.', 'lmaxs.', 'sctol.', 'dinit.',  &
                                 'dtinit', 'd0init', 'dfac..', 'dltfdc',  &
                                 'dltfdj', 'delta0', 'fuzz..', 'rlimit',  &
                                 'cosmin', 'huberc', 'rsptol', 'sigmin',  &
                                 'eta0..', 'bias..' /)

REAL (dp) :: vm(34) = (/ 1.0D-3, -0.99D+0, 1.0D-3, 1.0D-2, 1.2D+0, 1.0d-2,   &
                         1.2D+0, 0.d+0, 0.d+0, 1.d-3, -1.d+0, 0.d+0, 0.d+0,  &
                         0.d+0, 0.d+0, 0.d+0, 0.d+0, 0.d+0, 0.d+0, -10.d+0,  &
                         0.d+0, 0.d+0, 0.d+0, 0.d+0, 0.d+0, 0.d+0, 1.01D+0,  &
                         1.d+10, 0.d+0, 0.d+0, 0.d+0, 0.d+0, 0.d+0, 0.d+0 /)

REAL (dp) :: vx(34) = (/ 0.9D+0, -1.d-3, 1.d+1, 0.8D+0, 1.d+2, 0.8D+0,      &
                         1.d+2, 0.5D+0, 0.5D+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0, &
                         0.1D+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0,  &
                         1.d+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0, 1.d+10,  &
                         1.d+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0, 1.d+0 /)

CHARACTER (LEN=1) :: varnm(2)  = (/ 'p', 'n' /), sh(2) = (/ 's', 'h' /)
CHARACTER (LEN=4) :: cngd(3) = (/ '---c','hang','ed v' /),  &
                     dflt(3) = (/'nond','efau','lt v' /)

INTEGER :: ijmp=33, jlim(2) = (/ 0, 24 /), ndflt(2) = (/ 32, 25 /)
INTEGER :: miniv(2) = (/ 80, 59 /)

!...............................  body  ................................

pu=0
IF (prunit <= liv) pu=iv(prunit)
IF (alg < 1 .OR. alg > 2) GO TO 340
IF (iv(1) == 0) CALL deflt (alg,iv,liv,lv,v)
iv1=iv(1)
IF (iv1 /= 13 .AND. iv1 /= 12) GO TO 10
miv1=miniv(alg)
IF (perm <= liv) miv1=MAX(miv1,iv(perm)-1)
IF (ivneed <= liv) miv2=miv1+MAX(iv(ivneed),0)
IF (lastiv <= liv) iv(lastiv)=miv2
IF (liv < miv1) GO TO 300
iv(ivneed)=0
iv(lastv)=MAX(iv(vneed),0)+iv(lmat)-1
iv(vneed)=0
IF (liv < miv2) GO TO 300
IF (lv < iv(lastv)) GO TO 320
10 IF (alg == iv(algsav)) GO TO 30
IF (pu /= 0) WRITE (pu,20) alg,iv(algsav)
20 FORMAT (/' the first PARAMETER TO deflt should be',i3,' rather than',I3)
iv(1)=82
GO TO 360
30 IF (iv1 < 12 .OR. iv1 > 14) GO TO 60
IF (n >= 1) GO TO 50
iv(1)=81
IF (pu == 0) GO TO 360
WRITE (pu,40) varnm(alg),n
40 FORMAT (/' /// bad',a1,' =',i5)
GO TO 360
50 IF (iv1 /= 14) iv(nextiv)=iv(perm)
IF (iv1 /= 14) iv(nextv)=iv(lmat)
IF (iv1 == 13) GO TO 360
k=iv(parsav)-epslon
CALL vdflt (alg,v(k+1:))
iv(dtype0)=2-alg
iv(oldn)=n
which(1)=dflt(1)
which(2)=dflt(2)
which(3)=dflt(3)
GO TO 110
60 IF (n == iv(oldn)) GO TO 80
iv(1)=17
IF (pu == 0) GO TO 360
WRITE (pu,70) varnm(alg),iv(oldn),n
70 FORMAT (/' /// ',1A1,' changed from ',i5,' TO ',i5)
GO TO 360

80 IF (iv1 <= 11 .AND. iv1 >= 1) GO TO 100
iv(1)=80
IF (pu /= 0) WRITE (pu,90) iv1
90 FORMAT (/' ///  iv(1) =',i5,' should be between 0 AND 14.')
GO TO 360

100 which(1)=cngd(1)
which(2)=cngd(2)
which(3)=cngd(3)

110 IF (iv1 == 14) iv1=12
IF (big > tiny) GO TO 120
tiny=rmdcon(1)
machep=rmdcon(3)
big=rmdcon(6)
vm(12)=machep
vx(12)=big
vx(13)=big
vm(14)=machep
vm(17)=tiny
vx(17)=big
vm(18)=tiny
vx(18)=big
vx(20)=big
vx(21)=big
vx(22)=big
vm(24)=machep
vm(25)=machep
vm(26)=machep
vx(28)=rmdcon(5)
vm(29)=machep
vx(30)=big
vm(33)=machep
120 m=0
i=1
j=jlim(alg)
k=epslon
ndfalt=ndflt(alg)
DO l=1,ndfalt
  vk=v(k)
  IF (vk >= vm(i) .AND. vk <= vx(i)) GO TO 140
  m=k
  IF (pu /= 0) WRITE (pu,130) vn(i),k,vk,vm(i),vx(i)
  130 FORMAT (/' ///  ', A6, '.. v(', i2, ') =', f11.3, ' should',  &
              ' be between', f11.3, ' AND', f11.3)
  140 k=k + 1
  i=i + 1
  IF (i == j) i=ijmp
END DO

IF (iv(nvdflt) == ndfalt) GO TO 170
iv(1)=51
IF (pu == 0) GO TO 360
WRITE (pu,160) iv(nvdflt),ndfalt
160 FORMAT (/' iv(nvdflt) =',i5,' rather than ',i5)
GO TO 360
170 IF ((iv(dtype) > 0 .OR. v(dinit) > zero) .AND. iv1 == 12) GO TO 200
DO i=1,n
  IF (d(i) > zero) CYCLE
  m=18
  IF (pu /= 0) WRITE (pu,180) i,d(i)
  180 FORMAT (/' ///  d(',i3,') =',f11.3,' should be positive')
END DO
200 IF (m == 0) GO TO 210
iv(1)=m
GO TO 360

210 IF (pu == 0 .OR. iv(parprt) == 0) GO TO 360
IF (iv1 /= 12 .OR. iv(inits) == alg-1) GO TO 230
m=1
WRITE (pu,220) sh(alg),iv(inits)
220 FORMAT (/' nondefault values....'/' init',a1,'..... iv(25) =',i3)
230 IF (iv(dtype) == iv(dtype0)) GO TO 250
IF (m == 0) WRITE (pu,260) which
m=1
WRITE (pu,240) iv(dtype)
240 FORMAT (' dtype..... iv(16) =',i3)
250 i=1
j=jlim(alg)
k=epslon
l=iv(parsav)
ndfalt=ndflt(alg)
DO ii=1,ndfalt
  IF (v(k) == v(l)) GO TO 280
  IF (m == 0) WRITE (pu,260) which
  260 FORMAT (/' ',3A4,'alues....'/)
  m=1
  WRITE (pu,270) vn(i),k,v(k)
  270 FORMAT (' ',A6,'.. v(',i2,') =',f15.7)
  280 k=k + 1
  l=l + 1
  i=i + 1
  IF (i == j) i=ijmp
END DO

iv(dtype0)=iv(dtype)
parsv1=iv(parsav)
CALL vcopy (iv(nvdflt),v(parsv1:),v(epslon:))
GO TO 360

300 iv(1)=15
IF (pu == 0) GO TO 360
WRITE (pu,310) liv,miv2
310 FORMAT (/' /// liv =',i5,' must be at least',i5)
IF (liv < miv1) GO TO 360
IF (lv < iv(lastv)) GO TO 320
GO TO 360

320 iv(1)=16
IF (pu == 0) GO TO 360
WRITE (pu,330) lv,iv(lastv)
330 FORMAT (/' /// lv =',i5,' must be at least',i5)
GO TO 360

340 iv(1)=67
IF (pu == 0) GO TO 360
WRITE (pu,350) alg
350 FORMAT (/' /// alg =',i5,' must be 1 OR 2')

360 RETURN
!  ***  last card of parck follows  ***
END SUBROUTINE parck


FUNCTION reldst (p,d,x,x0) RESULT(fn_val)

!  ***  compute and return relative difference between x and x0 ***
!  ***  nl2sol version 2.2  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)   :: p
REAL (dp), INTENT(IN) :: d(:),x(:),x0(:)
REAL (dp)             :: fn_val

INTEGER   :: i
REAL (dp) :: emax,t,xmax,zero=0._dp

emax=zero
xmax=zero
DO i=1,p
  t=ABS(d(i)*(x(i)-x0(i)))
  IF (emax < t) emax=t
  t=d(i)*(ABS(x(i)) + ABS(x0(i)))
  IF (xmax < t) xmax=t
END DO
fn_val=zero
IF (xmax > zero) fn_val=emax/xmax
RETURN
!  ***  last card of reldst follows  ***
END FUNCTION reldst


FUNCTION stopx() RESULT(stopped)
!     *****parameters...
USE machine_constants
IMPLICIT NONE
LOGICAL :: stopped

!     ..................................................................

!     *****purpose...
!     this function may serve as the stopx (asynchronous interruption)
!     function for the nl2sol (nonlinear least-squares) package at
!     those installations which do not wish to implement a dynamic stopx.

!     *****algorithm notes...
!     at installations where the nl2sol system is used
!     interactively, this dummy stopx should be replaced by a
!     function that returns .true. if and only if the interrupt
!     (break) key has been pressed since the last call on stopx.

!     ..................................................................

stopped=.false.
RETURN
END FUNCTION stopx


SUBROUTINE vaxpy (p,w,a,x,y)

!  ***  set w = a*x + y  --  w, x, y = p-vectors, a = scalar  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: p
REAL (dp), INTENT(IN)  :: a,x(:),y(:)
REAL (dp), INTENT(OUT) :: w(:)

w(1:p)=a*x(1:p) + y(1:p)

RETURN
END SUBROUTINE vaxpy


SUBROUTINE vcopy (p,y,x)

!  ***  set y = x, where x and y are p-vectors  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: p
REAL (dp), INTENT(IN)  :: x(:)
REAL (dp), INTENT(OUT) :: y(:)

y(1:p)=x(1:p)

RETURN
END SUBROUTINE vcopy


SUBROUTINE vdflt (alg,v)

!  ***  supply ***sol (version 2.3) default values to v  ***

!  ***  alg = 1 means regression constants.
!  ***  alg = 2 means general unconstrained optimization constants.

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: alg
REAL (dp), INTENT(OUT) :: v(:)

REAL (dp) :: machep,mepcrt,one=1._dp,sqteps,three=3._dp

!  ***  v subscript values  ***

INTEGER   :: afctol=31, bias=43, cosmin=47, decfac=22, delta0=44, dfac=41,  &
             dinit=38, dltfdc=42, dltfdj=43, dtinit=39, d0init=40, epslon=19, &
             eta0=42, fuzz=45, huberc=48, incfac=23, lmax0=35, lmaxs=36,   &
             phmnfc=20, phmxfc=21, rdfcmn=24, rdfcmx=25, rfctol=32,   &
             rlimit=46, rsptol=49, sctol=37, sigmin=50, tuner1=26, tuner2=27, &
             tuner3=28, tuner4=29, tuner5=30, xctol=33, xftol=34

!-------------------------------  body  --------------------------------

machep=rmdcon(3)
v(afctol)=1.d-20
IF (machep > 1.d-10) v(afctol)=machep**2
v(decfac)=0.5D+0
sqteps=rmdcon(4)
v(dfac)=0.6D+0
v(delta0)=sqteps
v(dtinit)=1.d-6
mepcrt=machep**(one/three)
v(d0init)=1.d+0
v(epslon)=0.1D+0
v(incfac)=2.d+0
v(lmax0)=1.d+0
v(lmaxs)=1.d+0
v(phmnfc)=-0.1D+0
v(phmxfc)=0.1D+0
v(rdfcmn)=0.1D+0
v(rdfcmx)=4.d+0
v(rfctol)=MAX(1.d-10,mepcrt**2)
v(sctol)=v(rfctol)
v(tuner1)=0.1D+0
v(tuner2)=1.d-4
v(tuner3)=0.75D+0
v(tuner4)=0.5D+0
v(tuner5)=0.75D+0
v(xctol)=sqteps
v(xftol)=1.d+2*machep

IF (alg >= 2) GO TO 10

!  ***  regression  values

v(cosmin)=MAX(1.d-6,1.d+2*machep)
v(dinit)=0.d+0
v(dltfdc)=mepcrt
v(dltfdj)=sqteps
v(fuzz)=1.5D+0
v(huberc)=0.7D+0
v(rlimit)=rmdcon(5)
v(rsptol)=1.d-3
v(sigmin)=1.d-4
GO TO 20

!  ***  general optimization values

10 v(bias)=0.8D+0
v(dinit)=-1.0D+0
v(eta0)=1.0D+3*machep

20 RETURN
!  ***  last card of vdflt follows  ***
END SUBROUTINE vdflt


SUBROUTINE vscopy (p,y,s)

!  ***  set p-vector y to scalar s  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: p
REAL (dp), INTENT(IN)  :: s
REAL (dp), INTENT(OUT) :: y(:)

y(1:p)=s

RETURN
END SUBROUTINE vscopy


FUNCTION v2norm (p,x) RESULT(fn_val)

!  ***  return the 2-norm of the p-vector x, taking  ***
!  ***  care to avoid the most likely underflows.    ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)   :: p
REAL (dp), INTENT(IN) :: x(:)
REAL (dp)             :: fn_val

INTEGER   :: i,j
REAL (dp) :: one=1._dp,r,scale,sqteta=0._dp,t,xi,zero=0._dp

IF (p > 0) GO TO 10
fn_val=zero
GO TO 70
10 DO i=1,p
  IF (x(i) /= zero) GO TO 30
END DO
fn_val=zero
GO TO 70

30 scale=ABS(x(i))
IF (i < p) GO TO 40
fn_val=scale
GO TO 70
40 t=one
IF (sqteta == zero) sqteta=rmdcon(2)

!     ***  sqteta is (slightly larger than) the square root of the
!     ***  smallest positive floating point number on the machine.
!     ***  the tests involving sqteta are done to prevent underflows.

j=i + 1
DO i=j,p
  xi=ABS(x(i))
  IF (xi > scale) GO TO 50
  r=xi/scale
  IF (r > sqteta) t=t + r*r
  CYCLE
  50 r=scale/xi
  IF (r <= sqteta) r=zero
  t=one + t*r*r
  scale=xi
END DO

fn_val=scale*SQRT(t)
70 RETURN
!  ***  last card of v2norm follows  ***
END FUNCTION v2norm


SUBROUTINE humsl (n,d,x,calcf,calcgh,iv,liv,lv,v,uiparm,urparm,ufparm)

!  ***  minimize general unconstrained objective function using   ***
!  ***  (analytic) gradient and hessian provided by the caller.   ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: liv,lv,n
INTEGER, INTENT(IN OUT)   :: iv(:),uiparm(:)
REAL (dp), INTENT(IN OUT) :: d(:),x(:),v(:)
REAL (dp), INTENT(IN OUT) :: urparm(:),ufparm

!     dimension v(78 + n*(n+12)), uiparm(*), urparm(*)

INTERFACE
  SUBROUTINE calcf(n, x, nf, f, uiparm, urparm, ufparm)
    USE machine_constants
    IMPLICIT NONE
    INTEGER, INTENT(IN)       :: n
    INTEGER, INTENT(IN OUT)   :: nf, uiparm(:)
    REAL (dp), INTENT(IN)     :: x(:)
    REAL (dp), INTENT(OUT)    :: f, urparm(:), ufparm
  END SUBROUTINE calcf

  SUBROUTINE calcgh(n, x, nf, g, h, uiparm, urparm, ufparm)
    USE machine_constants
    IMPLICIT NONE
    INTEGER, INTENT(IN)       :: n, uiparm(:)
    INTEGER, INTENT(IN OUT)   :: nf
    REAL (dp), INTENT(IN)     :: x(:), ufparm
    REAL (dp), INTENT(IN OUT) :: urparm(:)
    REAL (dp), INTENT(OUT)    :: g(:), h(:)
  END SUBROUTINE calcgh
END INTERFACE

!------------------------------  discussion  ---------------------------

!        this routine is like sumsl, except that the subroutine para-
!     meter calcg of sumsl (which computes the gradient of the objec-
!     tive function) is replaced by the subroutine parameter calcgh,
!     which computes both the gradient and (lower triangle of the)
!     hessian of the objective function.  the calling sequence is...
!             call calcgh(n, x, nf, g, h, uiparm, urparm, ufparm)
!     parameters n, x, nf, g, uiparm, urparm, and ufparm are the same
!     as for sumsl, while h is an array of length n*(n+1)/2 in which
!     calcgh must store the lower triangle of the hessian at x.  start-
!     ing at h(1), calcgh must store the hessian entries in the order
!     (1,1), (2,1), (2,2), (3,1), (3,2), (3,3), ...
!        the value printed (by itsum) in the column labelled stppar
!     is the levenberg-marquardt used in computing the current step.
!     zero means a full newton step.  if the special case described in
!     ref. 1 is detected, then stppar is negated.  the value printed
!     in the column labelled npreldf is zero if the current hessian
!     is not positive definite.
!        it sometimes proves worthwhile to let d be determined from the
!     diagonal of the hessian matrix by setting iv(dtype) = 1 and
!     v(dinit) = 0.  the following iv and v components are relevant...

! iv(dtol)..... iv(59) gives the starting subscript in v of the dtol
!             array used when d is updated.  (iv(dtol) can be
!             initialized by calling humsl with iv(1) = 13.)
! iv(dtype).... iv(16) tells how the scale vector d should be chosen.
!             iv(dtype) .le. 0 means that d should not be updated, and
!             iv(dtype) .ge. 1 means that d should be updated as
!             described below with v(dfac).  default = 0.
! v(dfac)..... v(41) and the dtol and d0 arrays (see v(dtinit) and
!             v(d0init)) are used in updating the scale vector d when
!             iv(dtype) .gt. 0.  (d is initialized according to
!             v(dinit), described in sumsl.)  let
!                  d1(i) = max(sqrt(abs(h(i,i))), v(dfac)*d(i)),
!             where h(i,i) is the i-th diagonal element of the current
!             hessian.  if iv(dtype) = 1, then d(i) is set to d1(i)
!             unless d1(i) .lt. dtol(i), in which case d(i) is set to
!                  max(d0(i), dtol(i)).
!             if iv(dtype) .ge. 2, then d is updated during the first
!             iteration as for iv(dtype) = 1 (after any initialization
!             due to v(dinit)) and is left unchanged thereafter.
!             default = 0.6.
! v(dtinit)... v(39), if positive, is the value to which all components
!             of the dtol array (see v(dfac)) are initialized.  if
!             v(dtinit) = 0, then it is assumed that the caller has
!             stored dtol in v starting at v(iv(dtol)).
!             default = 10**-6.
! v(d0init)... v(40), if positive, is the value to which all components
!             of the d0 vector (see v(dfac)) are initialized.  if
!             v(dfac) = 0, then it is assumed that the caller has
!             stored d0 in v starting at v(iv(dtol)+n).  default = 1.0.

!  ***  reference  ***

! 1. gay, d.m. (1981), computing optimal locally constrained steps,
!         siam j. sci. statist. comput. 2, pp. 186-197.
!.
!  ***  general  ***

!     coded by david m. gay (winter 1980).  revised sept. 1982.
!     this subroutine was written in connection with research supported
!     in part by the national science foundation under grants
!     mcs-7600324 and mcs-7906671.

!----------------------------  declarations  ---------------------------

! deflt... provides default input values for iv and v.
! humit... reverse-communication routine that does humsl algorithm.

INTEGER   :: g1,h1,iv1,lh,nf
REAL (dp) :: f

!  ***  subscripts for iv   ***

INTEGER   :: nextv=47, nfcall=6, nfgcal=7, g=28, h=56, toobig=2, vneed=4

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

lh=n*(n+1)/2
IF (iv(1) == 0) CALL deflt (2,iv,liv,lv,v)
IF (iv(1) == 12 .OR. iv(1) == 13) iv(vneed)=iv(vneed) + n*(n+3)/2
iv1=iv(1)
IF (iv1 == 14) GO TO 10
IF (iv1 > 2 .AND. iv1 < 12) GO TO 10
g1=1
h1=1
IF (iv1 == 12) iv(1)=13
GO TO 20

10 g1=iv(g)
h1=iv(h)

20 CALL humit (d,f,v(g1:),v(h1:),iv,lh,liv,lv,n,v,x)
IF (iv(1)-2 < 0) THEN
  GO TO 30
ELSE IF (iv(1)-2 == 0) THEN
  GO TO 40
ELSE
  GO TO 50
END IF

30 nf=iv(nfcall)
CALL calcf (n,x,nf,f,uiparm,urparm,ufparm)
IF (nf <= 0) iv(toobig)=1
GO TO 20

40 CALL calcgh (n,x,iv(nfgcal),v(g1:),v(h1:),uiparm,urparm,ufparm)
GO TO 20

50 IF (iv(1) /= 14) GO TO 60

!  ***  storage allocation

iv(g)=iv(nextv)
iv(h)=iv(g) + n
iv(nextv)=iv(h) + n*(n+1)/2
IF (iv1 /= 13) GO TO 10

60 RETURN
!  ***  last card of humsl follows  ***
END SUBROUTINE humsl


SUBROUTINE humit (d,fx,g,h,iv,lh,liv,lv,n,v,x)

!  ***  carry out humsl (unconstrained minimization) iterations, using
!  ***  hessian matrix provided by the caller.

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: lh,liv,lv,n
INTEGER, INTENT(IN OUT)   :: iv(:)
REAL (dp), INTENT(IN OUT) :: d(:),fx,g(:),h(:),v(:),x(:)

!--------------------------  parameter usage  --------------------------

! d.... scale vector.
! fx... function value.
! g.... gradient vector.
! h.... lower triangle of the hessian, stored rowwise.
! iv... integer value array.
! lh... length of h = p*(p+1)/2.
! liv.. length of iv (at least 60).
! lv... length of v (at least 78 + n*(n+21)/2).
! n.... number of variables (components in x and g).
! v.... floating-point value array.
! x.... parameter vector.

!  ***  discussion  ***

!        parameters iv, n, v, and x are the same as the corresponding
!     ones to humsl (which see), except that v can be shorter (since
!     the part of v that humsl uses for storing g and h is not needed).
!     moreover, compared with humsl, iv(1) may have the two additional
!     output values 1 and 2, which are explained below, as is the use
!     of iv(toobig) and iv(nfgcal).  the value iv(g), which is an
!     output value from humsl, is not referenced by humit or the
!     subroutines it calls.

! iv(1) = 1 means the caller should set fx to f(x), the function value
!             at x, and call humit again, having changed none of the
!             other parameters.  an exception occurs if f(x) cannot be
!             computed (e.g. if overflow would occur), which may happen
!             because of an oversized step.  in this case the caller
!             should set iv(toobig) = iv(2) to 1, which will cause
!             humit to ignore fx and try a smaller step.  the para-
!             meter nf that humsl passes to calcf (for possible use by
!             calcgh) is a copy of iv(nfcall) = iv(6).
! iv(1) = 2 means the caller should set g to g(x), the gradient of f at
!             x, and h to the lower triangle of h(x), the hessian of f
!             at x, and call humit again, having changed none of the
!             other parameters except perhaps the scale vector d.
!                  the parameter nf that humsl passes to calcg is
!             iv(nfgcal) = iv(7).  if g(x) and h(x) cannot be evaluated,
!             then the caller may set iv(nfgcal) to 0, in which case
!             humit will return with iv(1) = 65.
!                  note -- humit overwrites h with the lower triangle
!             of  diag(d)**-1 * h(x) * diag(d)**-1.
!.
!  ***  general  ***

!     coded by david m. gay (winter 1980).  revised sept. 1982.
!     this subroutine was written in connection with research supported
!     in part by the national science foundation under grants
!     mcs-7600324 and mcs-7906671.

!        (see sumsl and humsl for references.)

!+++++++++++++++++++++++++++  declarations  ++++++++++++++++++++++++++++

!  ***  local variables  ***

INTEGER   :: dg1,i,j,k,l,lstgst,nn1o2,step1,temp1,w1,x01
REAL (dp) :: t

!  ***  external functions and subroutines  ***

! assst.... assesses candidate step.
! deflt.... provides default iv and v input values.
! dotprd... returns inner product of two vectors.
! dupdu.... updates scale vector d.
! gqtst.... computes optimally locally constrained step.
! itsum.... prints iteration summary and info on initial and final x.
! parck.... checks validity of input iv and v values.
! reldst... computes v(reldx) = relative step size.
! slvmul... multiplies symmetric matrix times vector, given the lower
!             triangle of the matrix.
! stopx.... returns .true. if the break key has been pressed.
! vaxpy.... computes scalar times one vector plus another.
! vcopy.... copies one vector to another.
! vscopy... sets all elements of a vector to a scalar.
! v2norm... returns the 2-norm of a vector.

!  ***  iv subscript values  ***

INTEGER :: cnvcod=55, dg=37, dtol=59, dtype=16, irc=29, kagqt=33, lmat=42,  &
           mode=35, model=5, mxfcal=17, mxiter=18, nextv=47, nfcall=6,   &
           nfgcal=7, ngcall=30, niter=31, radinc=8, restor=9, step=40,   &
           stglim=11, stlstg=41, toobig=2, vneed=4, w=34, xirc=13, x0=43

!  ***  v subscript values  ***

INTEGER :: dgnorm=1, dinit=38, dstnrm=2, dtinit=39, d0init=40, f=10, f0=13,  &
           fdif=11, gtstep=4, incfac=23, lmax0=35, lmaxs=36, preduc=7,   &
           radfac=16, radius=8, rad0=9, reldx=17, stppar=5, tuner4=29,   &
           tuner5=30

REAL (dp) :: one=1.d+0, onep2=1.2D+0, zero=0.d+0

!+++++++++++++++++++++++++++++++  body  ++++++++++++++++++++++++++++++++

i=iv(1)
IF (i == 1) GO TO 30
IF (i == 2) GO TO 40

!  ***  check validity of iv and v input values  ***

IF (iv(1) == 0) CALL deflt (2,iv,liv,lv,v)
IF (iv(1) == 12 .OR. iv(1) == 13) iv(vneed)=iv(vneed) + n*(n+21)/2+7
CALL parck (2,d,iv,liv,lv,n,v)
i=iv(1)-2
IF (i > 12) GO TO 360
nn1o2=n*(n+1)/2
IF (lh < nn1o2) THEN
  iv(1)=66
  GO TO 350
END IF

SELECT CASE (i)
  CASE (1:6)
    GO TO 210
  CASE (7)
    GO TO 160
  CASE (8)
    GO TO 120
  CASE (9)
    GO TO 160
  CASE (10:11)
    GO TO 10
  CASE DEFAULT
    GO TO 20
END SELECT

!  ***  storage allocation  ***

10 iv(dtol)=iv(lmat) + nn1o2
iv(x0)=iv(dtol) + 2*n
iv(step)=iv(x0) + n
iv(stlstg)=iv(step) + n
iv(dg)=iv(stlstg) + n
iv(w)=iv(dg) + n
iv(nextv)=iv(w) + 4*n + 7
IF (iv(1) /= 13) GO TO 20
iv(1)=14
GO TO 360

!  ***  initialization  ***

20 iv(niter)=0
iv(nfcall)=1
iv(ngcall)=1
iv(nfgcal)=1
iv(mode)=-1
iv(model)=1
iv(stglim)=1
iv(toobig)=0
iv(cnvcod)=0
iv(radinc)=0
v(rad0)=zero
v(stppar)=zero
IF (v(dinit) >= zero) CALL vscopy (n,d,v(dinit))
k=iv(dtol)
IF (v(dtinit) > zero) CALL vscopy (n,v(k:),v(dtinit))
k=k + n
IF (v(d0init) > zero) CALL vscopy (n,v(k:),v(d0init))
iv(1)=1
GO TO 360

30 v(f)=fx
IF (iv(mode) >= 0) GO TO 210
iv(1)=2
IF (iv(toobig) == 0) GO TO 360
iv(1)=63
GO TO 350

!  ***  make sure gradient could be computed  ***

40 IF (iv(nfgcal) /= 0) GO TO 50
iv(1)=65
GO TO 350

!  ***  update the scale vector d  ***

50 dg1=iv(dg)
IF (iv(dtype) <= 0) GO TO 70
k=dg1
j=0
DO i=1,n
  j=j + i
  v(k)=h(j)
  k=k + 1
END DO
CALL dupdu (d,v(dg1:),iv,n,v)

!  ***  compute scaled gradient and its norm  ***

70 dg1=iv(dg)
k=dg1
DO i=1,n
  v(k)=g(i)/d(i)
  k=k + 1
END DO
v(dgnorm)=v2norm(n,v(dg1:))

!  ***  compute scaled hessian  ***

k=1
DO i=1,n
  t=one/d(i)
  DO j=1,i
    h(k)=t*h(k)/d(j)
    k=k + 1
  END DO
END DO

IF (iv(cnvcod) /= 0) GO TO 340
IF (iv(mode) == 0) GO TO 300

!  ***  allow first step to have scaled 2-norm at most v(lmax0)  ***

v(radius)=v(lmax0)

iv(mode)=0


!-----------------------------  main loop  -----------------------------


!  ***  print iteration summary, check iteration limit  ***

110 CALL itsum (d,g,iv,n,v,x)
120 k=iv(niter)
IF (k < iv(mxiter)) GO TO 130
iv(1)=10
GO TO 350

130 iv(niter)=k + 1

!  ***  initialize for start of next iteration  ***

dg1=iv(dg)
x01=iv(x0)
v(f0)=v(f)
iv(irc)=4
iv(kagqt)=-1

!     ***  copy x to x0 ***

CALL vcopy (n,v(x01:),x)

!  ***  update radius  ***

IF (k == 0) GO TO 150
step1=iv(step)
k=step1
DO i=1,n
  v(k)=d(i)*v(k)
  k=k + 1
END DO
v(radius)=v(radfac)*v2norm(n,v(step1:))

!  ***  check stopx and function evaluation limit  ***

150 IF (.NOT.stopx()) GO TO 170
iv(1)=11
GO TO 180

!     ***  come here when restarting after func. eval. limit or stopx.

160 IF (v(f) >= v(f0)) GO TO 170
v(radfac)=one
k=iv(niter)
GO TO 130

170 IF (iv(nfcall) < iv(mxfcal)) GO TO 190
iv(1)=9
180 IF (v(f) >= v(f0)) GO TO 350

!        ***  in case of stopx or function evaluation limit with
!        ***  improved v(f), evaluate the gradient at x.

iv(cnvcod)=iv(1)
GO TO 290

!. . . . . . . . . . . . .  compute candidate step  . . . . . . . . . .

190 step1=iv(step)
dg1=iv(dg)
l=iv(lmat)
w1=iv(w)
CALL gqtst (d,v(dg1:),h,iv(kagqt),v(l:),n,v(step1:),v,v(w1:))
IF (iv(irc) == 6) GO TO 210

!  ***  check whether evaluating f(x0 + step) looks worthwhile  ***

IF (v(dstnrm) <= zero) GO TO 210
IF (iv(irc) /= 5) GO TO 200
IF (v(radfac) <= one) GO TO 200
IF (v(preduc) <= onep2*v(fdif)) GO TO 210

!  ***  compute f(x0 + step)  ***

200 x01=iv(x0)
step1=iv(step)
CALL vaxpy (n,x,one,v(step1:),v(x01:))
iv(nfcall)=iv(nfcall) + 1
iv(1)=1
iv(toobig)=0
GO TO 360

!. . . . . . . . . . . . .  assess candidate step  . . . . . . . . . . .

210 x01=iv(x0)
v(reldx)=reldst(n,d,x,v(x01:))
CALL assst (iv,v)
step1=iv(step)
lstgst=iv(stlstg)
IF (iv(restor) == 1) CALL vcopy (n,x,v(x01:))
IF (iv(restor) == 2) CALL vcopy (n,v(lstgst:),v(step1:))
IF (iv(restor) /= 3) GO TO 220
CALL vcopy (n,v(step1:),v(lstgst:))
CALL vaxpy (n,x,one,v(step1:),v(x01:))
v(reldx)=reldst(n,d,x,v(x01:))

220 k=iv(irc)
SELECT CASE (k)
  CASE (1, 5)      !     ***  recompute step with new radius  ***
    v(radius)=v(radfac)*v(dstnrm)
    GO TO 150
  CASE (2:4)       !. . . . . . . . . . . .  process acceptable step  . . . . . . . . . . .
    IF (iv(irc) /= 3) GO TO 290
    temp1=lstgst

    !     ***  prepare for gradient tests  ***
    !     ***  set  temp1 = hessian * step + g(x0)
    !     ***             = diag(d) * (h * step + g(x0))

    !        use x0 vector as temporary.
    k=x01
    DO i=1,n
      v(k)=d(i)*v(step1)
      k=k + 1
      step1=step1 + 1
    END DO
    CALL slvmul (n,v(temp1:),h,v(x01:))
    DO i=1,n
      v(temp1)=d(i)*v(temp1) + g(i)
      temp1=temp1 + 1
    END DO
  CASE (6)         !  ***  compute step of length v(lmaxs) for singular convergence test.
    v(radius)=v(lmaxs)
    GO TO 190
  CASE (7:12)      !  ***  convergence or false convergence  ***
    iv(cnvcod)=k-4
    IF (v(f) >= v(f0)) GO TO 340
    IF (iv(xirc) == 14) GO TO 340
    iv(xirc)=14
  CASE (13)
    GO TO 330
  CASE (14)
    GO TO 300
END SELECT

!  ***  compute gradient and hessian  ***

290 iv(ngcall)=iv(ngcall) + 1
iv(1)=2
GO TO 360

300 iv(1)=2
IF (iv(irc) /= 3) GO TO 110

!  ***  set v(radfac) by gradient tests  ***

temp1=iv(stlstg)
step1=iv(step)

!     ***  set  temp1 = diag(d)**-1 * (hessian*step + (g(x0)-g(x)))  ***

k=temp1
DO i=1,n
  v(k)=(v(k)-g(i))/d(i)
  k=k + 1
END DO

!     ***  do gradient tests  ***

IF (v2norm(n,v(temp1:)) <= v(dgnorm)*v(tuner4)) GO TO 320
IF (dotprd(n,g,v(step1:)) >= v(gtstep)*v(tuner5)) GO TO 110
320 v(radfac)=v(incfac)
GO TO 110

!. . . . . . . . . . . . . .  misc. details  . . . . . . . . . . . . . .

!  ***  bad parameters to assess  ***

330 iv(1)=64
GO TO 350

!  ***  print summary of final iteration and other requested items  ***

340 iv(1)=iv(cnvcod)
iv(cnvcod)=0
350 CALL itsum (d,g,iv,n,v,x)

360 RETURN

!  ***  last card of humit follows  ***
END SUBROUTINE humit


SUBROUTINE dupdu (d,hdiag,iv,n,v)

!  ***  update scale vector d for humsl  ***

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: n,iv(:)
REAL (dp), INTENT(IN)  :: hdiag(:),v(:)
REAL (dp), INTENT(OUT) :: d(:)

!  ***  local variables  ***

INTEGER   :: dtoli,d0i,i
REAL (dp) :: t,vdfac

!  ***  subscripts for iv and v  ***

INTEGER :: dfac = 41, dtol = 59, dtype = 16, niter = 31

!-------------------------------  body  --------------------------------

i=iv(dtype)
IF (i == 1) GO TO 10
IF (iv(niter) > 0) GO TO 30

10 dtoli=iv(dtol)
d0i=dtoli + n
vdfac=v(dfac)
DO i=1,n
  t=MAX(SQRT(ABS(hdiag(i))),vdfac*d(i))
  IF (t < v(dtoli)) t=MAX(v(dtoli),v(d0i))
  d(i)=t
  dtoli=dtoli + 1
  d0i=d0i + 1
END DO

30 RETURN
!  ***  last card of dupdu follows  ***
END SUBROUTINE dupdu


SUBROUTINE gqtst (d,dig,dihdi,ka,l,p,step,v,w)

!  *** compute goldfeld-quandt-trotter step by more-hebden technique ***
!  ***  (nl2sol version 2.2), modified a la more and sorensen  ***

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: p
INTEGER, INTENT(IN OUT)   :: ka
REAL (dp), INTENT(IN)     :: d(:),dig(:)
REAL (dp), INTENT(IN OUT) :: l(:),v(:),step(:),w(:),dihdi(:)
!     dimension dihdi(p*(p+1)/2), l(p*(p+1)/2), w(4*p+7)

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

!  ***  purpose  ***

!        given the (compactly stored) lower triangle of a scaled
!     hessian (approximation) and a nonzero scaled gradient vector,
!     this subroutine computes a goldfeld-quandt-trotter step of
!     approximate length v(radius) by the more-hebden technique.  in
!     other words, step is computed to (approximately) minimize
!     psi(step) = (g**t)*step + 0.5*(step**t)*h*step  such that the
!     2-norm of d*step is at most (approximately) v(radius), where
!     g  is the gradient,  h  is the hessian, and  d  is a diagonal
!     scale matrix whose diagonal is stored in the parameter d.
!     (gqtst assumes  dig = d**-1 * g  and  dihdi = d**-1 * h * d**-1.)

!  ***  parameter description  ***

!     d (in)  = the scale vector, i.e. the diagonal of the scale
!              matrix  d  mentioned above under purpose.
!   dig (in)  = the scaled gradient vector, d**-1 * g.  if g = 0, then
!              step = 0 and  v(stppar) = 0 are returned.
! dihdi (in)  = lower triangle of the scaled hessian (approximation),
!              i.e., d**-1 * h * d**-1, stored compactly by rows., i.e.,
!              in the order (1,1), (2,1), (2,2), (3,1), (3,2), etc.
!    ka (i/o) = the number of hebden iterations (so far) taken to deter-
!              mine step.  ka .lt. 0 on input means this is the first
!              attempt to determine step (for the present dig and dihdi)
!              -- ka is initialized to 0 in this case.  output with
!              ka = 0 (or v(stppar) = 0)  means  step = -(h**-1)*g.
!     l (i/o) = workspace of length p*(p+1)/2 for cholesky factors.
!     p (in)  = number of parameters -- the hessian is a  p x p  matrix.
!  step (i/o) = the step computed.
!     v (i/o) contains various constants and variables described below.
!     w (i/o) = workspace of length 4*p + 6.

!  ***  entries in v  ***

! v(dgnorm) (i/o) = 2-norm of (d**-1)*g.
! v(dstnrm) (output) = 2-norm of d*step.
! v(dst0)   (i/o) = 2-norm of d*(h**-1)*g (for pos. def. h only), or
!             overestimate of smallest eigenvalue of (d**-1)*h*(d**-1).
! v(epslon) (in)  = max. rel. error allowed for psi(step).  for the
!             step returned, psi(step) will exceed its optimal value
!             by less than -v(epslon)*psi(step).  suggested value = 0.1.
! v(gtstep) (out) = inner product between g and step.
! v(nreduc) (out) = psi(-(h**-1)*g) = psi(newton step)  (for pos. def.
!             h only -- v(nreduc) is set to zero otherwise).
! v(phmnfc) (in)  = tol. (together with v(phmxfc)) for accepting step
!             (more*s sigma).  the error v(dstnrm) - v(radius) must lie
!             between v(phmnfc)*v(radius) and v(phmxfc)*v(radius).
! v(phmxfc) (in)  (see v(phmnfc).)
!             suggested values -- v(phmnfc) = -0.25, v(phmxfc) = 0.5.
! v(preduc) (out) = psi(step) = predicted obj. func. reduction for step.
! v(radius) (in)  = radius of current (scaled) trust region.
! v(rad0)   (i/o) = value of v(radius) from previous call.
! v(stppar) (i/o) is normally the marquardt parameter, i.e. the alpha
!             described below under algorithm notes.  if h + alpha*d**2
!             (see algorithm notes) is (nearly) singular, however,
!             then v(stppar) = -alpha.

!  ***  usage notes  ***

!     if it is desired to recompute step using a different value of
!     v(radius), then this routine may be restarted by calling it
!     with all parameters unchanged except v(radius).  (this explains
!     why step and w are listed as i/o).  on an initial call (one with
!     ka .lt. 0), step and w need not be initialized and only compo-
!     nents v(epslon), v(stppar), v(phmnfc), v(phmxfc), v(radius), and
!     v(rad0) of v must be initialized.

!  ***  algorithm notes  ***

!        the desired g-q-t step (ref. 2, 3, 4, 6) satisfies
!     (h + alpha*d**2)*step = -g  for some nonnegative alpha such that
!     h + alpha*d**2 is positive semidefinite.  alpha and step are
!     computed by a scheme analogous to the one described in ref. 5.
!     estimates of the smallest and largest eigenvalues of the hessian
!     are obtained from the gerschgorin circle theorem enhanced by a
!     simple form of the scaling described in ref. 7.  cases in which
!     h + alpha*d**2 is nearly (or exactly) singular are handled by
!     the technique discussed in ref. 2.  in these cases, a step of
!     (exact) length v(radius) is returned for which psi(step) exceeds
!     its optimal value by less than -v(epslon)*psi(step).  the test
!     suggested in ref. 6 for detecting the special case is performed
!     once two matrix factorizations have been done -- doing so sooner
!     seems to degrade the performance of optimization routines that
!     call this routine.

!  ***  functions and subroutines called  ***

! dotprd - returns inner product of two vectors.
! litvmu - applies inverse-transpose of compact lower triang. matrix.
! livmul - applies inverse of compact lower triang. matrix.
! lsqrt  - finds cholesky factor (of compactly stored lower triang.).
! lsvmin - returns approx. to min. sing. value of lower triang. matrix.
! rmdcon - returns machine-dependent constants.
! v2norm - returns 2-norm of a vector.

!  ***  references  ***

! 1.  dennis, j.e., gay, d.m., and welsch, r.e. (1981), an adaptive
!             nonlinear least-squares algorithm, acm trans. math.
!             software, vol. 7, no. 3.
! 2.  gay, d.m. (1981), computing optimal locally constrained steps,
!             siam j. sci. statist. computing, vol. 2, no. 2, pp.
!             186-197.
! 3.  goldfeld, s.m., quandt, r.e., and trotter, h.f. (1966),
!             maximization by quadratic hill-climbing, econometrica 34,
!             pp. 541-551.
! 4.  hebden, m.d. (1973), an algorithm for minimization using exact
!             second derivatives, report t.p. 515, theoretical physics
!             div., a.e.r.e. harwell, oxon., england.
! 5.  more, j.j. (1978), the levenberg-marquardt algorithm, implemen-
!             tation and theory, pp.105-116 of springer lecture notes
!             in mathematics no. 630, edited by g.a. watson, springer-
!             verlag, berlin and new york.
! 6.  more, j.j., and sorensen, d.c. (1981), computing a trust region
!             step, technical report anl-81-83, argonne national lab.
! 7.  varga, r.s. (1965), minimal gerschgorin sets, pacific j. math. 15,
!             pp. 719-729.

!  ***  general  ***

!     coded by david m. gay.
!     this subroutine was written in connection with research
!     supported by the national science foundation under grants
!     mcs-7600324, dcr75-10143, 76-14311dss, mcs76-11989, and
!     mcs-7906671.

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

!  ***  local variables  ***

LOGICAL   :: restrt
INTEGER   :: dggdmx,diag,diag0,dstsav,emax,emin,i,im1,inc,irc,j,k,  &
             kalim,kamin,k1,lk0,phipin,q,q0,uk0,x
REAL (dp) :: alphak,aki,akk,delta,dst,eps,gtsta,lk,oldphi,phi,  &
             phimax,phimin,psifac,rad,radsq,root,si,sk,sw,t,twopsi,t1,t2,uk,wi

!  ***  subscripts for v  ***

INTEGER :: dgnorm=1, dstnrm=2, dst0=3, epslon=19, gtstep=4, nreduc=6,   &
           phmnfc=20, phmxfc=21, preduc=7, radius=8, rad0=9, stppar=5

REAL (dp) :: epsfac=50.0D+0, four=4.0D+0, half=0.5D+0, kappa=2.0D+0,   &
             negone=-1.0D+0, one=1.0D+0, p001=1.0D-3, six=6.0D+0,   &
             three=3.0D+0, two=2.0D+0, zero=0.0D+0
REAL (dp) :: big=0.d+0, dgxfac=0.d+0

!  ***  body  ***

!     ***  store largest abs. entry in (d**-1)*h*(d**-1) at w(dggdmx).
dggdmx=p + 1
!     ***  store gerschgorin over- and underestimates of the largest
!     ***  and smallest eigenvalues of (d**-1)*h*(d**-1) at w(emax)
!     ***  and w(emin) respectively.
emax=dggdmx + 1
emin=emax + 1
!     ***  for use in recomputing step, the final values of lk, uk, dst,
!     ***  and the inverse derivative of more*s phi at 0 (for pos. def.
!     ***  h) are stored in w(lk0), w(uk0), w(dstsav), and w(phipin)
!     ***  respectively.
lk0=emin + 1
phipin=lk0 + 1
uk0=phipin + 1
dstsav=uk0 + 1
!     ***  store diag of (d**-1)*h*(d**-1) in w(diag),...,w(diag0+p).
diag0=dstsav
diag=diag0 + 1
!     ***  store -d*step in w(q),...,w(q0+p).
q0=diag0 + p
q=q0 + 1
!     ***  allocate storage for scratch vector x  ***
x=q + p
rad=v(radius)
radsq=rad**2
!     ***  phitol = max. error allowed in dst = v(dstnrm) = 2-norm of
!     ***  d*step.
phimax=v(phmxfc)*rad
phimin=v(phmnfc)*rad
psifac=two*v(epslon)/(three*(four*(v(phmnfc) + one)*(kappa + one) +   &
kappa + two)*rad**2)
!     ***  oldphi is used to detect limits of numerical accuracy.  if
!     ***  we recompute step and it does not change, then we accept it.
oldphi=zero
eps=v(epslon)
irc=0
restrt=.false.
kalim=ka + 50

!  ***  start or restart, depending on ka  ***

IF (ka >= 0) GO TO 290

!  ***  fresh start  ***

k=0
uk=negone
ka=0
kalim=50
v(dgnorm)=v2norm(p,dig)
v(nreduc)=zero
v(dst0)=zero
kamin=3
IF (v(dgnorm) == zero) kamin=0

!     ***  store diag(dihdi) in w(diag0+1),...,w(diag0+p)  ***

j=0
DO i=1,p
  j=j + i
  k1=diag0 + i
  w(k1)=dihdi(j)
END DO

!     ***  determine w(dggdmx), the largest element of dihdi  ***

t1=zero
j=p*(p+1)/2
DO i=1,j
  t=ABS(dihdi(i))
  IF (t1 < t) t1=t
END DO
w(dggdmx)=t1

!  ***  try alpha = 0 ***

30 CALL lsqrt (1,p,l,dihdi,irc)
IF (irc == 0) GO TO 50
!        ***  indef. h -- underestimate smallest eigenvalue, use this
!        ***  estimate to initialize lower bound lk on alpha.
j=irc*(irc+1)/2
t=l(j)
l(j)=one
DO i=1,irc
  w(i)=zero
END DO
w(irc)=one
CALL litvmu (irc,w,l,w)
t1=v2norm(irc,w)
lk=-t/t1/t1
v(dst0)=-lk
IF (restrt) GO TO 210
GO TO 70

!     ***  positive definite h -- compute unmodified newton step.  ***
50 lk=zero
CALL lsvmin(p,l,w(q:),w(q:),t)
IF (t >= one) GO TO 60
IF (big <= zero) big=rmdcon(6)
IF (v(dgnorm) >= t*t*big) GO TO 70
60 CALL livmul (p,w(q:),l,dig)
gtsta=dotprd(p,w(q:),w(q:))
v(nreduc)=half*gtsta
CALL litvmu (p,w(q:),l,w(q:))
dst=v2norm(p,w(q:))
v(dst0)=dst
phi=dst-rad
IF (phi <= phimax) GO TO 260
IF (restrt) GO TO 210

!  ***  prepare to compute gerschgorin estimates of largest (and
!  ***  smallest) eigenvalues.  ***

70 k=0
DO i=1,p
  wi=zero
  IF (i == 1) GO TO 90
  im1=i-1
  DO j=1,im1
    k=k + 1
    t=ABS(dihdi(k))
    wi=wi + t
    w(j)=w(j) + t
  END DO
  90 w(i)=wi
  k=k + 1
END DO

!  ***  (under-)estimate smallest eigenvalue of (d**-1)*h*(d**-1)  ***

k=1
t1=w(diag)-w(1)
IF (p <= 1) GO TO 120
DO i=2,p
  j=diag0 + i
  t=w(j)-w(i)
  IF (t >= t1) CYCLE
  t1=t
  k=i
END DO

120 sk=w(k)
j=diag0 + k
akk=w(j)
k1=k*(k-1)/2 + 1
inc=1
t=zero
DO i=1,p
  IF (i == k) GO TO 130
  aki=ABS(dihdi(k1))
  si=w(i)
  j=diag0 + i
  t1=half*(akk-w(j) + si-aki)
  t1=t1 + SQRT(t1*t1 + sk*aki)
  IF (t < t1) t=t1
  IF (i < k) GO TO 140
  130 inc=i
  140 k1=k1 + inc
END DO

w(emin)=akk-t
uk=v(dgnorm)/rad-w(emin)
IF (v(dgnorm) == zero) uk=uk + p001 + p001*uk
IF (uk <= zero) uk=p001

!  ***  compute gerschgorin (over-)estimate of largest eigenvalue  ***

k=1
t1=w(diag) + w(1)
IF (p <= 1) GO TO 170
DO i=2,p
  j=diag0 + i
  t=w(j) + w(i)
  IF (t <= t1) CYCLE
  t1=t
  k=i
END DO

170 sk=w(k)
j=diag0 + k
akk=w(j)
k1=k*(k-1)/2 + 1
inc=1
t=zero
DO i=1,p
  IF (i == k) GO TO 180
  aki=ABS(dihdi(k1))
  si=w(i)
  j=diag0 + i
  t1=half*(w(j) + si-aki-akk)
  t1=t1 + SQRT(t1*t1 + sk*aki)
  IF (t < t1) t=t1
  IF (i < k) GO TO 190
  180 inc=i
  190 k1=k1 + inc
END DO

w(emax)=akk + t
lk=MAX(lk,v(dgnorm)/rad-w(emax))

!     ***  alphak = current value of alpha (see alg. notes above).  we
!     ***  use more*s scheme for initializing it.
alphak=ABS(v(stppar))*v(rad0)/rad

IF (irc /= 0) GO TO 210

!  ***  compute l0 for positive definite h  ***

CALL livmul (p,w,l,w(q:))
t=v2norm(p,w)
w(phipin)=dst/t/t
lk=MAX(lk,phi*w(phipin))

!  ***  safeguard alphak and add alphak*i to (d**-1)*h*(d**-1)  ***

210 ka=ka + 1
IF (-v(dst0) >= alphak .OR. alphak < lk .OR. alphak >= uk)  &
                        alphak=uk* MAX(p001,SQRT(lk/uk))
IF (alphak <= zero) alphak=half*uk
IF (alphak <= zero) alphak=uk
k=0
DO i=1,p
  k=k + i
  j=diag0 + i
  dihdi(k)=w(j) + alphak
END DO

!  ***  try computing cholesky decomposition  ***

CALL lsqrt (1,p,l,dihdi,irc)
IF (irc == 0) GO TO 240

!  ***  (d**-1)*h*(d**-1) + alphak*i  is indefinite -- overestimate
!  ***  smallest eigenvalue for use in updating lk  ***

j=(irc*(irc+1))/2
t=l(j)
l(j)=one
DO i=1,irc
  w(i)=zero
END DO
w(irc)=one
CALL litvmu (irc,w,l,w)
t1=v2norm(irc,w)
lk=alphak-t/t1/t1
v(dst0)=-lk
GO TO 210

!  ***  alphak makes (d**-1)*h*(d**-1) positive definite.
!  ***  compute q = -d*step, check for convergence.  ***

240 CALL livmul (p,w(q:),l,dig)
gtsta=dotprd(p,w(q:),w(q:))
CALL litvmu (p,w(q:),l,w(q:))
dst=v2norm(p,w(q:))
phi=dst-rad
IF (phi <= phimax .AND. phi >= phimin) GO TO 270
IF (phi == oldphi) GO TO 270
oldphi=phi
IF (phi < zero) GO TO 330

!  ***  unacceptable alphak -- update lk, uk, alphak  ***

250 IF (ka >= kalim) GO TO 270
!     ***  the following MIN is necessary because of restarts  ***
IF (phi < zero) uk=MIN(uk,alphak)
!     *** kamin = 0 only iff the gradient vanishes  ***
IF (kamin == 0) GO TO 210
CALL livmul (p,w,l,w(q:))
t1=v2norm(p,w)
alphak=alphak + (phi/t1)*(dst/t1)*(dst/rad)
lk=MAX(lk,alphak)
GO TO 210

!  ***  acceptable step on first try  ***

260 alphak=zero

!  ***  successful step in general.  compute step = -(d**-1)*q  ***

270 DO i=1,p
  j=q0 + i
  step(i)=-w(j)/d(i)
END DO
v(gtstep)=-gtsta
v(preduc)=half*(ABS(alphak)*dst*dst + gtsta)
GO TO 410


!  ***  restart with new radius  ***

290 IF (v(dst0) <= zero .OR. v(dst0)-rad > phimax) GO TO 310

!     ***  prepare to return newton step  ***

restrt=.true.
ka=ka + 1
k=0
DO i=1,p
  k=k + i
  j=diag0 + i
  dihdi(k)=w(j)
END DO
uk=negone
GO TO 30

310 kamin=ka + 3
IF (v(dgnorm) == zero) kamin=0
IF (ka == 0) GO TO 50

dst=w(dstsav)
alphak=ABS(v(stppar))
phi=dst-rad
t=v(dgnorm)/rad
uk=t-w(emin)
IF (v(dgnorm) == zero) uk=uk + p001 + p001*uk
IF (uk <= zero) uk=p001
IF (rad > v(rad0)) GO TO 320

!        ***  smaller radius  ***
lk=zero
IF (alphak > zero) lk=w(lk0)
lk=MAX(lk,t-w(emax))
IF (v(dst0) > zero) lk=MAX(lk,(v(dst0)-rad)*w(phipin))
GO TO 250

!     ***  bigger radius  ***
320 IF (alphak > zero) uk=MIN(uk,w(uk0))
lk=MAX(zero,-v(dst0),t-w(emax))
IF (v(dst0) > zero) lk=MAX(lk,(v(dst0)-rad)*w(phipin))
GO TO 250

!  ***  decide whether to check for special case... in practice (from
!  ***  the standpoint of the calling optimization code) it seems best
!  ***  not to check until a few iterations have failed -- hence the
!  ***  test on kamin below.

330 delta=alphak + MIN(zero,v(dst0))
twopsi=alphak*dst*dst + gtsta
IF (ka >= kamin) GO TO 340
!     *** if the test in ref. 2 is satisfied, fall through to handle
!     *** the special case (as soon as the more-sorensen test detects
!     *** it).
IF (delta >= psifac*twopsi) GO TO 370

!  ***  check for the special case of  h + alpha*d**2  (nearly)
!  ***  singular.  use one step of inverse power method with start
!  ***  from lsvmin to obtain approximate eigenvector corresponding
!  ***  to smallest eigenvalue of (d**-1)*h*(d**-1).  lsvmin returns
!  ***  x and w with  l*w = x.

340 CALL lsvmin(p,l,w(x:),w,t)

!     ***  normalize w  ***
w(1:p)=t*w(1:p)
!     ***  complete current inv. power iter. -- replace w by (l**-t)*w.
CALL litvmu (p,w,l,w)
t2=one/v2norm(p,w)
w(1:p)=t2*w(1:p)
t=t2*t

!  ***  now w is the desired approximate (unit) eigenvector and
!  ***  t*x = ((d**-1)*h*(d**-1) + alphak*i)*w.

sw=dotprd(p,w(q:),w)
t1=(rad + dst)*(rad-dst)
root=SQRT(sw*sw + t1)
IF (sw < zero) root=-root
si=t1/(sw + root)

!  ***  the actual test for the special case...

IF ((t2*si)**2 <= eps*(dst**2 + alphak*radsq)) GO TO 380

!  ***  update upper bound on smallest eigenvalue (when not positive)
!  ***  (as recommended by more and sorensen) and continue...

IF (v(dst0) <= zero) v(dst0)=MIN(v(dst0),t2**2-alphak)
lk=MAX(lk,-v(dst0))

!  ***  check whether we can hope to detect the special case in
!  ***  the available arithmetic.  accept step as it is if not.

!     ***  if not yet available, obtain machine dependent value dgxfac.
370 IF (dgxfac == zero) dgxfac=epsfac*rmdcon(3)

IF (delta > dgxfac*w(dggdmx)) GO TO 250
GO TO 270

!  ***  special case detected... negate alphak to indicate special case

380 alphak=-alphak
v(preduc)=half*twopsi

!  ***  accept current step if adding si*w would lead to a
!  ***  further relative reduction in psi of less than v(epslon)/3.

t1=zero
t=si*(alphak*sw-half*si*(alphak + t*dotprd(p,w(x:),w)))
IF (t < eps*twopsi/six) GO TO 390
v(preduc)=v(preduc) + t
dst=rad
t1=-si
390 DO i=1,p
  j=q0 + i
  w(j)=t1*w(i)-w(j)
  step(i)=w(j)/d(i)
END DO
v(gtstep)=dotprd(p,dig,w(q:))

!  ***  save values for use in a possible restart  ***

410 v(dstnrm)=dst
v(stppar)=alphak
w(lk0)=lk
w(uk0)=uk
v(rad0)=rad
w(dstsav)=dst

!     ***  restore diagonal of dihdi  ***

j=0
DO i=1,p
  j=j + i
  k=diag0 + i
  dihdi(j)=w(k)
END DO

RETURN

!  ***  last card of gqtst follows  ***
END SUBROUTINE gqtst


SUBROUTINE lsqrt (n1,n,l,a,irc)

!  ***  compute rows n1 through n of the cholesky factor  l  of
!  ***  a = l*(l**t),  where  l  and the lower triangle of  a  are both
!  ***  stored compactly by rows (and may occupy the same storage).
!  ***  irc = 0 means all went well.  irc = j means the leading
!  ***  principal  j x j  submatrix of  a  is not positive definite --
!  ***  and  l(j*(j+1)/2)  contains the (nonpos.) reduced j-th diagonal.

!  ***  parameters  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)       :: n1,n
INTEGER, INTENT(OUT)      :: irc
REAL (dp), INTENT(IN OUT) :: l(:)
REAL (dp), INTENT(IN)     :: a(:)

!     dimension l(n*(n+1)/2), a(n*(n+1)/2)

!  ***  local variables  ***

INTEGER   :: i,ij,ik,im1,i0,j,jk,jm1,j0,k
REAL (dp) :: t,td,zero=0._dp

!  ***  body  ***

i0=n1*(n1-1)/2
DO i=n1,n
  td=zero
  IF (i == 1) GO TO 40
  j0=0
  im1=i-1
  DO j=1,im1
    t=zero
    IF (j == 1) GO TO 20
    jm1=j-1
    DO k=1,jm1
      ik=i0 + k
      jk=j0 + k
      t=t + l(ik)*l(jk)
    END DO
    20 ij=i0 + j
    j0=j0 + j
    t=(a(ij)-t)/l(j0)
    l(ij)=t
    td=td + t*t
  END DO
  40 i0=i0 + i
  t=a(i0)-td
  IF (t <= zero) GO TO 60
  l(i0)=SQRT(t)
END DO

irc=0
GO TO 70

60 l(i0)=t
irc=i

70 RETURN

!  ***  last card of lsqrt  ***
END SUBROUTINE lsqrt


SUBROUTINE lsvmin (p,l,x,y,fn_val)
! N.B. Original function changed to a subroutine for compatibility with ELF90

!  ***  estimate smallest singular value of packed lower triangular matrix L

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: p
REAL (dp), INTENT(IN)  :: l(:)
REAL (dp), INTENT(OUT) :: x(:),y(:),fn_val
!     dimension l(p*(p+1)/2)

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

!  ***  purpose  ***

!     this subroutine returns a good over-estimate of the smallest
!     singular value of the packed lower triangular matrix l.

!  ***  parameter description  ***

!  p (in)  = the order of l.  l is a  p x p  lower triangular matrix.
!  l (in)  = array holding the elements of  l  in row order, i.e.
!             l(1,1), l(2,1), l(2,2), l(3,1), l(3,2), l(3,3), etc.
!  x (out) if lsvmin returns a positive value, then x is a normalized
!             approximate left singular vector corresponding to the
!             smallest singular value.  this approximation may be very
!             crude.  if lsvmin returns zero, then some components of x
!             are zero and the rest retain their input values.
!  y (out) if lsvmin returns a positive value, then y = (l**-1)*x is an
!             unnormalized approximate right singular vector correspond-
!             ing to the smallest singular value.  this approximation
!             may be crude.  if lsvmin returns zero, then y retains its
!             input value.  the caller may pass the same vector for x
!             and y (nonstandard fortran usage), in which case y over-
!             writes x (for nonzero lsvmin returns).

!  ***  algorithm notes  ***

!     the algorithm is based on (1), with the additional provision that
!     lsvmin = 0 is returned if the smallest diagonal element of l
!     (in magnitude) is not more than the unit roundoff times the
!     largest.  the algorithm uses a random number generator proposed
!     in (4), which passes the spectral test with flying colors -- see
!     (2) and (3).

!  ***  subroutines and functions called  ***

!        v2norm - function, returns the 2-norm of a vector.

!  ***  references  ***

!     (1) cline, a., moler, c., stewart, g., and wilkinson, j.h.(1977),
!         an estimate for the condition number of a matrix, report
!         tm-310, applied math. div., argonne national laboratory.

!     (2) hoaglin, d.c. (1976), theoretical properties of congruential
!         random-number generators --  an empirical view,
!         memorandum ns-340, dept. of statistics, harvard univ.

!     (3) knuth, d.e. (1969), the art of computer programming, vol. 2
!         (seminumerical algorithms), addison-wesley, reading, mass.

!     (4) smith, c.s. (1971), multiplicative pseudo-random number
!         generators with prime modulus, j. assoc. comput. mach. 18,
!         pp. 586-593.

!  ***  history  ***

!     designed and coded by david m. gay (winter 1977/summer 1978).

!  ***  general  ***

!     this subroutine was written in connection with research
!     supported by the national science foundation under grants
!     mcs-7600324, dcr75-10143, 76-14311dss, and mcs76-11989.

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

!  ***  local variables  ***

INTEGER   :: i,ii,ix,j,ji,jj,jjj,jm1,j0,pm1
REAL (dp) :: b,sminus,splus,t,xminus,xplus

!  ***  constants  ***

REAL (dp) :: half=0.5D+0, one=1.d+0, r9973=9973.d+0, zero=0.d+0

!  ***  body  ***

ix=2
pm1=p-1

!  ***  first check whether to return lsvmin = 0 and initialize x  ***

ii=0
j0=p*pm1/2
jj=j0 + p
IF (l(jj) == zero) GO TO 80
ix=MOD(3432*ix,9973)
b=half*(one + REAL(ix)/r9973)
xplus=b/l(jj)
x(p)=xplus
IF (p <= 1) GO TO 50
DO i=1,pm1
  ii=ii + i
  IF (l(ii) == zero) GO TO 80
  ji=j0 + i
  x(i)=xplus*l(ji)
END DO

!  ***  solve (l**t)*x = b, where the components of b have randomly
!  ***  chosen magnitudes in (.5,1) with signs chosen to make x large.

!     do j = p-1 to 1 by -1...
DO jjj=1,pm1
  j=p-jjj
!       ***  determine x(j) in this iteration. note for i = 1,2,...,j
!       ***  that x(i) holds the current partial sum for row i.
  ix=MOD(3432*ix,9973)
  b=half*(one + REAL(ix)/r9973)
  xplus=(b-x(j))
  xminus=(-b-x(j))
  splus=ABS(xplus)
  sminus=ABS(xminus)
  jm1=j-1
  j0=j*jm1/2
  jj=j0 + j
  xplus=xplus/l(jj)
  xminus=xminus/l(jj)
  IF (jm1 == 0) GO TO 30
  DO i=1,jm1
    ji=j0 + i
    splus=splus + ABS(x(i) + l(ji)*xplus)
    sminus=sminus + ABS(x(i) + l(ji)*xminus)
  END DO
  30 IF (sminus > splus) xplus=xminus
  x(j)=xplus
!       ***  update partial sums  ***
  IF (jm1 > 0) CALL vaxpy (jm1,x,xplus,l(j0+1:),x)
END DO

!  ***  normalize x  ***

50 t=one/v2norm(p,x)
x(1:p)=t*x(1:p)

!  ***  solve l*y = x and return lsvmin = 1/twonorm(y)  ***

DO j=1,p
  jm1=j-1
  j0=j*jm1/2
  jj=j0 + j
  t=zero
  IF (jm1 > 0) t=dotprd(jm1,l(j0+1:),y)
  y(j)=(x(j)-t)/l(jj)
END DO

fn_val=one/v2norm(p,y)
GO TO 90

80 fn_val=zero
90 RETURN
!  ***  last card of lsvmin follows  ***
END SUBROUTINE lsvmin


SUBROUTINE slvmul (p,y,s,x)

!  ***  set  y = s * x,  s = p x p symmetric matrix.  ***
!  ***  lower triangle of  s  stored rowwise.         ***

!  ***  parameter declarations  ***

USE machine_constants
IMPLICIT NONE
INTEGER, INTENT(IN)    :: p
REAL (dp), INTENT(IN)  :: s(:),x(:)
REAL (dp), INTENT(OUT) :: y(:)
!     dimension s(p*(p+1)/2)

!  ***  local variables  ***

INTEGER   :: i,im1,j,k
REAL (dp) :: xi

!-----------------------------------------------------------------------

j=1
DO i=1,p
  y(i)=dotprd(i,s(j:),x)
  j=j + i
END DO

IF (p <= 1) GO TO 40
j=1
DO i=2,p
  xi=x(i)
  im1=i-1
  j=j + 1
  DO k=1,im1
    y(k)=y(k) + s(j)*xi
    j=j + 1
  END DO
END DO

40 RETURN
!  ***  last card of slvmul follows  ***
END SUBROUTINE slvmul

END MODULE toms611
